# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import base64
import dataclasses
import json
import uuid
from dataclasses import field
from enum import IntEnum
from pathlib import Path
from typing import Any, Protocol, TypeVar, runtime_checkable

from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import BaseConfig

from ..interfaces import AnyPacketType


class MsgTypes(IntEnum):
    """
    Messages used to support initialization and configuration of MCTP communication within an MCTP network, as
    specified in DSP0236
    """

    CTRL = 0x00
    PLDM = 0x01
    NCSI = 0x02
    Ethernet = 0x03
    NVMeMgmtMsg = 0x04
    SPDM = 0x05
    SECUREDMSG = 0x06
    CXL_FM_API = 0x07
    CXL_CCI = 0x08
    VDPCI = 0x7E
    VDIANA = 0x7F


class VendorIdFormat(IntEnum):
    PCI_VENDOR_ID = 0x0
    """16-bit Unsigned Integer that identifies the manufacturer
    of the device. Valid vendor identifiers are allocated by the
    PCI SIG to ensure uniqueness."""

    IANA_ENT_NUMBER = 0x1
    """The IANA enterprise number for the organization or vendor
    expressed as a 32-bit unsigned binary number."""


@dataclasses.dataclass(frozen=True)
class VendorCapabilitySet(DataClassDictMixin):
    vendor_id: int
    id_format: VendorIdFormat = VendorIdFormat.PCI_VENDOR_ID
    command_set_type: int = 0


class PhysicalTransportBindingId(IntEnum):
    Reserved = 0x0
    MCTPoverSMBus = 0x01
    MCTPoverPcieVdm = 0x02
    MCTPoverKCS = 0x04
    MCTPoverSerial = 0x05
    MCTPoverI3C = 0x06
    VendorDefined = 0xFF


class PhysicalMediumIdentifiers(IntEnum):
    Unspecified = 0x0
    SMBUS_2_0_100khz = 0x1
    SMBUS_2_0_I2C_100khz = 0x2
    I2C_100khz = 0x3
    SMBUS_3_0_I2C_400khz = 0x4
    SMBUS_3_0_I2C_1mhz = 0x5
    I2C_3_4mhz = 0x6
    PCIeRev_1_1 = 0x8
    PCIeRev_2_0 = 0x9
    PCIeRev_2_1 = 0xA
    PCIeRev_3 = 0xB
    PCIeRev_4 = 0xC
    PCIeRev_5 = 0xD
    PCICompatible = 0xF
    USB_1_1 = 0x10
    USB_2 = 0x11
    USB_3 = 0x12
    NCSIOverRBT = 0x18
    KCSLegacy = 0x20
    KCSPCI = 0x21
    SerialHostLegacy = 0x22
    SerialHostPCI = 0x23
    AsyncSerial = 0x24
    I3CBasic = 0x30


class EntryType(IntEnum):
    SINGLE_ENDPOINT = 0
    BRIDGE_AND_DOWNSTREAM_ENDPOINT = 1
    SINGLE_BRIDGE_ENDPOINT = 2
    ADDITIONAL_BRIDGE_EID_RANGE = 3


@dataclasses.dataclass(frozen=True)
class RoutingTableEntry(DataClassDictMixin):
    starting_eid: int
    port_number: int
    phy_address: list[int]
    phys_transport_binding_id: PhysicalTransportBindingId = PhysicalTransportBindingId.MCTPoverSMBus
    phy_media_type_id: PhysicalMediumIdentifiers = PhysicalMediumIdentifiers.I2C_100khz
    entry_type: EntryType = EntryType.SINGLE_ENDPOINT
    eid_range: int = 1
    static_eid: bool = False


@dataclasses.dataclass(frozen=True)
class Smbus7bitAddress(DataClassDictMixin):
    address: int

    def __post_init__(self):
        if self.address > 0x7F:
            msg = f"SMBUS address {hex(self.address)} is more than 7 bits."
            raise ValueError(msg)

    def read(self) -> int:
        return self.address << 1 | 0x01

    def write(self) -> int:
        return self.address << 1 | 0x00


@dataclasses.dataclass(frozen=True)
class Smbus10bitAddress(DataClassDictMixin):
    address: int

    def __post_init__(self):
        if self.address > 0x3FF:
            msg = f"SMBUS address {hex(self.address)} is more than 10 bits."
            raise ValueError(msg)


AnyPhysicalAddress = TypeVar("AnyPhysicalAddress", Smbus7bitAddress, Smbus10bitAddress)


@dataclasses.dataclass(frozen=True)
class MctpResponse(DataClassDictMixin):
    request: list[int]
    response: list[int]
    processing_delay: int = field(metadata=field_options(alias="processing-delay"))
    key: bytes = field(init=False)
    data: bytes = field(init=False)
    response_index: int = 0
    description: str = "<missing>"

    def __post_init__(self):
        req = bytes(self.request)
        key = base64.b64encode(req)
        object.__setattr__(self, "key", key)

        packet = bytes(self.response)
        object.__setattr__(self, "data", packet)


# @dataclasses.dataclass(frozen=True)
# class VdPciResponeList(DataClassDictMixin):
#     vendors: Dict[str, Dict[str, List[MctpResponse]]]


@dataclasses.dataclass(frozen=True)
class MctpResponseList(DataClassDictMixin):
    responses: dict[MsgTypes, list[MctpResponse] | dict[str, dict[str, list[MctpResponse]]]]


def deserialize_msg_types(values: list[MsgTypes] | list[str]) -> list[MsgTypes]:
    if not isinstance(values, list):
        msg = f"Unknown msg type: {type(values)}"
        raise TypeError(msg)
    if all(isinstance(y, MsgTypes) for y in values):
        return values
    msg_types = []
    for mt in values:
        if isinstance(mt, str) and hasattr(MsgTypes, mt):
            msg_types += [MsgTypes[mt]]
            continue
        msg = f"Unknown msg type: {mt}"
        raise TypeError(msg)
    return msg_types


@dataclasses.dataclass
class EndpointContext(DataClassDictMixin):
    physical_address: Smbus7bitAddress
    static_eid: int | None = None
    assigned_eid: int = 0
    discovered: bool = False
    is_bus_owner: bool = False
    pool_size: int = 0
    mtu_size: int = 240
    allocated_pool: list[int] | None = None
    endpoint_uuid: uuid.UUID = dataclasses.field(default_factory=lambda: uuid.uuid4())
    supported_msg_types: list[MsgTypes] = dataclasses.field(default_factory=lambda: [MsgTypes.CTRL])
    supported_vdm_msg_types: list[VendorCapabilitySet] = dataclasses.field(default_factory=list)
    mctp_responses: MctpResponseList | None = None
    routing_table_ready: bool = False
    routing_table: list[RoutingTableEntry] = dataclasses.field(default_factory=list)

    class Config(BaseConfig):
        serialization_strategy = {list[MsgTypes]: {"deserialize": deserialize_msg_types}}

    @property
    def eid(self) -> int:
        if self.assigned_eid:
            return self.assigned_eid
        if self.static_eid:
            return self.static_eid
        # no EID assigned, just return the null EID
        return 0

    def import_json_responses(self, resp_file: Path):
        with resp_file.open("r") as f:
            data: dict[str, Any] = json.load(f)
            data = {MsgTypes[k]: v for k, v in data.items() if MsgTypes[k] in self.supported_msg_types}
            self.mctp_responses = MctpResponseList.from_dict({"responses": data})

    def get_response(
        self, msg_type: MsgTypes, req_bytes: bytes, vendor_id: str = "", vdm_cmd_code: str = ""
    ) -> MctpResponse | None:
        if msg_type not in MsgTypes:
            return None
        if msg_type not in self.mctp_responses.responses:
            return None
        key = base64.b64encode(req_bytes)
        if msg_type != MsgTypes.VDPCI:
            for resp in self.mctp_responses.responses[msg_type]:
                if resp.key == key:
                    return resp
        else:
            responses = self.mctp_responses.responses[msg_type]
            if vendor_id not in responses:
                return None
            responses = responses[vendor_id]
            if vdm_cmd_code not in responses:
                return None
            for resp in responses[vdm_cmd_code]:
                if resp.key == key:
                    return resp
        return None


@runtime_checkable
class ICanReply(Protocol):
    def is_request(self, check_payload: bool = True) -> bool:
        pass

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        pass



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from typing import Protocol, TypeVar, runtime_checkable

from scapy.packet import Packet
from scapy.plist import PacketList

AnyPacketType = TypeVar("AnyPacketType", Packet, PacketList, None)


@runtime_checkable
class ICanSetMySummaryClasses(Protocol):
    def set_mysummary_classes(self, classes):
        pass


@runtime_checkable
class ICanVerifyIfRequest(Protocol):
    def is_request(self, check_payload: bool = True) -> bool:
        pass



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import contextlib
import threading
import time
from operator import itemgetter

from scapy.compat import raw
from scapy.data import MTU
from scapy.packet import Packet
from scapy.supersocket import SuperSocket
from scapy.utils import hexdump, linehexdump

from ..layers.mctp import Smbus7bitAddress, SmbusTransport

try:
    from array import array

    import pyaardvark
except RuntimeError:
    # ignore the missing library as this might not be needed in all deployments
    pyaardvark = None


class AardvarkI2CSocket(SuperSocket):
    desc = "read/write to an Aardvark USB device"

    def __init__(
        self,
        slave_address: Smbus7bitAddress,
        port: int | None = None,
        serial_number: str | None = None,
        enable_i2c_pullups: bool = False,
        enable_target_power: bool = False,
        slave_only: bool = False,
        poll_period_ms: int = 10,
        id_str: str = "",
        dump_hex=True,
        dump_packet=False,
        **kwargs,
    ):
        if pyaardvark is None:
            msg = "Failed to load pyaardvark library. Confirm if environment is bootstrapped."
            raise RuntimeError(msg)
        self.id_str = id_str
        self.dump_hex = dump_hex
        self.dump_packet = dump_packet
        self._slave_address = slave_address
        self._port = port
        self._serial_number = serial_number
        self._enable_i2c_pullups = enable_i2c_pullups
        self._enable_target_power = enable_target_power
        self._dev: pyaardvark.Aardvark | None = None
        self._lock = threading.Lock()
        self._slave_only = slave_only
        self._poll_period_ms = poll_period_ms

        if not self.connect():
            msg = f"Failed to open connection to Aardvark adapter: {slave_address}, {port}"
            raise RuntimeError(msg)

    def connect(self) -> bool:
        """
        Claims the Aardvark device defined when initializing class. This can be used to release the device
        for a short period of time and reconnect without having to recreate the surrounding objects
        (e.g. sessions, answering machines, ...).

        :return: True if claimed device, False if failed or already connected
        """
        if not self._dev or self._dev.handle is None:
            self._dev: pyaardvark.Aardvark = pyaardvark.open(self._port, self._serial_number)
            self._dev.enable_i2c_slave(self._slave_address.address)

            if self._enable_i2c_pullups:
                self._dev.i2c_pullups = self._enable_i2c_pullups
            if self._enable_target_power:
                self._dev.target_power = self._enable_target_power
            self._dev.i2c_stop()
            return True
        return False

    def close(self) -> None:
        """
        Releases the Aardvark device (to allow other applications to claim the device).
        :return: None
        """
        if self._dev:
            self._dev.close()

    def send(self, x: Packet) -> int:
        """
        Overloaded Packet.send() method to send data using Aardvark APIs
        """
        sx = raw(x)
        with contextlib.suppress(AttributeError):
            x.sent_time = time.time()

        if self.dump_hex:
            print(f"{self.id_str}>TX> {linehexdump(x, onlyhex=1, dump=True)}")
        if self.dump_packet:
            print(f"{self.id_str}>TX> {x.summary()}")

        try:
            with self._lock:
                # API uses 7bit addresses but payload has 8bit address in the first byte
                if self._slave_only:
                    self._dev.i2c_slave_response(sx[1:])
                    # TODO: wait until the msg is received
                else:
                    self._dev.i2c_master_write(sx[0] >> 1, sx[1:])
                    # this causes the stop condition to be skipped, hanging the bus
                    # which is great for testing error handling
                    # self._dev.i2c_master_write(sx[0] >> 1, sx[1:], flags=I2C_NO_STOP)
                    # time.sleep(0.5)
                    # self._dev.i2c_stop()
        except OSError as err:
            print(f"Aardvark write failed: {err}")
            code, *_ = err.args
            if code and code == pyaardvark.I2C_STATUS_SLA_NACK:
                return 0
                # the return value is not checked, just raise the exception
                # pass
            raise

        return len(sx)

    def recv(self, x: int = MTU) -> Packet | None:
        """
        Receives any pending data written to the slave address. Callers should first call "select()" to wait
        for data to be available to be received. This API call will sleep waiting for data if the buffer is
        empty.

        :param x: Ignored but part of overloaded methods signature
        :return: SmbusTransportPacket if available or None if no data is received
        """
        # TODO: read until there is no more data available (might not be needed as the library uses 64K read buffers)
        (i2c_addr, rx_data) = self._dev.i2c_slave_read()
        rx_data = array("B", rx_data)
        rq_sa = array(
            "B",
            [
                i2c_addr << 1,
            ],
        )
        raw_array = rq_sa + rx_data
        raw_bytes = raw_array.tobytes()

        if self.dump_hex:
            print(f"{self.id_str}<RX< {linehexdump(raw_bytes, onlyhex=1, dump=True)}")

        # Attempt to parse the packet but mask any parsing errors as no valid packets received
        pkt = None
        with contextlib.suppress(Exception):
            pkt = SmbusTransport(raw_bytes)
            pkt.time = time.time()

        if pkt and self.dump_packet:
            print(f"{self.id_str}<RX< {pkt.summary()}")

        return pkt

    @staticmethod
    def select(sockets: list[SuperSocket], remain: float | None = None) -> list[SuperSocket]:
        """
        This function is called during sendrecv() routine to select
        the available sockets.

        :param sockets: an array of sockets that need to be selected
        :param remain: remaining timeout (in seconds) to wait for data
        :returns: an array of sockets that were selected and
            the function to be called next to get the packets (i.g. recv)
        """
        aardvark_socks = [sock for sock in sockets if isinstance(sock, AardvarkI2CSocket)]
        if len(aardvark_socks) != 1:
            msg = "AardvarkI2C can only monitor a single socket at a time"
            raise RuntimeError(msg)
        self = aardvark_socks[0]

        # use "remain" if it is less than the pre-defined poll period
        events = self._dev.poll(int(min(self._poll_period_ms, (remain or 1) * 1000)))
        if pyaardvark.POLL_I2C_WRITE in events:
            transmit_size = self._dev.i2c_slave_last_transmit_size()
            print(f"DEBUG: last transmit size {transmit_size}")
        if pyaardvark.POLL_I2C_READ in events:
            return [self]
        if events and pyaardvark.POLL_I2C_WRITE not in events:
            print(f"DEBUG: events {events}")
        return []

    @staticmethod
    def show_devices() -> None:
        """
        Prints the Aardvark adapters present in the system to
        :return: None
        """
        devices = pyaardvark.find_devices()
        for dev in devices:
            port, serial_number, in_use = itemgetter("port", "serial_number", "in_use")(dev)
            print(f"{port}) {serial_number} [{'in-used' if in_use else 'free'}]")



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from .types import (
    CompletionCode,
    CompletionCodes,
    ContrlCmdCodes,
    IControlMsgCanReply,
    IControlMsgPacket,
)

from .control import (
    RqBit,
    ControlHdr,
    ControlHdrPacket,
)

from .get_eid import GetEndpointID, GetEndpointIDPacket, GetEndpointIDResponse, EndpointType, EndpointIDType

from .set_eid import (
    SetEndpointID,
    SetEndpointIDPacket,
    SetEndpointIDResponse,
    SetEndpointIDOperation,
    SetEndpointIDAssignmentStatus,
    SetEndpointIDAllocationStatus,
)

from .discovery_notify import (
    DiscoveryNotify,
    DiscoveryNotifyPacket,
    DiscoveryNotifyResponse,
)

from .get_eid_uuid import (
    GetEndpointUUID,
    GetEndpointUUIDPacket,
    GetEndpointUUIDResponse,
)

from .get_mctp_version_support import (
    GetMctpVersionSupport,
    GetMctpVersionSupportPacket,
    GetMctpVersionSupportResponse,
)

from .get_msg_type_support import (
    GetMessageTypeSupport,
    GetMessageTypeSupportPacket,
    GetMessageTypeSupportResponse,
)

from .get_vdm_support import (
    GetVendorDefinedMessageSupport,
    GetVendorDefinedMessageSupportPacket,
    GetVendorDefinedMessageSupportResponse,
    NO_MORE_CAPABILITY_SETS,
    VendorIdFormat,
)

from .allocate_eids import (
    AllocateEIDAllocationStatus,
    AllocateEIDOperation,
    AllocateEndpointIDs,
    AllocateEndpointIDsPacket,
    AllocateEndpointIDsResponse,
)

from .get_routing_table_entries import (
    EntryType,
    GetRoutingTableEntries,
    GetRoutingTableEntriesPacket,
    RoutingTableEntryPacket,
)



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from .vdpci import AutobindVDMMsg, RqBit, VdPciHdr, VdPciHdrPacket

from .types import VdPCIVendorIds



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict, defaultdict
from collections import OrderedDict as OrderedDictType
from pathlib import Path

from scapy.compat import raw
from scapy.utils import rdpcap

from .control import ControlHdrPacket
from .pldm import PldmControlCmdCodes, PldmHdrPacket, PldmPlatformMonitoringCmdCodes, PldmTypeCodes
from .transport import TransportHdrPacket
from .types import AnyPacketType, EndpointContext, MctpResponse, MctpResponseList, MsgTypes


def import_pcap_dump(resp_file: Path, endpoint_dump: bool, ctx: EndpointContext) -> MctpResponseList | None:
    if not resp_file.name.endswith(".dump") and not resp_file.name.endswith(".pcap"):
        return
    pending_reqs: list[AnyPacketType] = []
    responses: OrderedDictType[int, MctpResponse] = OrderedDict()
    responseList: dict[MsgTypes, list[MctpResponse]] = defaultdict(list)
    for resp_packet in rdpcap(str(resp_file.resolve())):
        # tx_packet = packet.pkttype == 4
        # prefix = '<TX<' if tx_packet else '>RX>'
        if not resp_packet.haslayer(TransportHdrPacket):
            continue
        packet = resp_packet.getlayer(TransportHdrPacket)
        # print(f"{prefix} {packet.summary()}")
        if (packet.haslayer(ControlHdrPacket) and packet.rq) or (packet.haslayer(PldmHdrPacket) and packet.rq):
            pending_reqs += [packet]
            continue

        # Assume this is a response and search for the request
        original_req: AnyPacketType = None
        for req in pending_reqs:
            if req.tag != packet.tag:
                continue
            if packet.to == req.to:
                continue
            if req.dst not in (packet.src, 0):
                continue
            original_req = req
            break
        else:
            pending_reqs += [packet]
            continue

        pending_reqs.remove(original_req)

        # TODO: move this code into the msg type packet layer by using an interface
        if packet.msg_type == MsgTypes.CTRL:
            req = raw(original_req.getlayer(ControlHdrPacket))[1:]
            rsp = raw(packet.getlayer(ControlHdrPacket))[1:]
            if req in responses:
                msg = "Found a duplicate request, stop and fix..."
                raise SystemExit(msg)
            mctp_resp = MctpResponse(
                request=list(req),
                response=list(rsp),
                processing_delay=0,
                description=original_req.getlayer(ControlHdrPacket).summary(),
            )
            responses[req] = mctp_resp
            responseList[MsgTypes.CTRL] += [mctp_resp]
        elif packet.msg_type == MsgTypes.PLDM:
            req = raw(original_req.getlayer(PldmHdrPacket))[1:]
            rsp = raw(packet.getlayer(PldmHdrPacket))[1:]
            if req in responses and responses[req].response == rsp:
                msg = "Found a duplicate request, stop and fix..."
                raise SystemExit(msg)
            type_code = PldmTypeCodes(original_req.pldm_type)
            if original_req.pldm_type == PldmTypeCodes.CONTROL:
                cmd_code_str = PldmControlCmdCodes(original_req.cmd_code).name
            elif original_req.pldm_type == PldmTypeCodes.PLATFORM_MONITORING:
                cmd_code_str = PldmPlatformMonitoringCmdCodes(original_req.cmd_code).name
            else:
                cmd_code_str = f"{original_req.cmd_code}({hex(original_req.cmd_code)})"
            mctp_resp = MctpResponse(
                request=list(req),
                response=list(rsp),
                processing_delay=0,
                description=f"PLDM {type_code.name} {cmd_code_str}",
            )
            responses[req] = mctp_resp
            responseList[MsgTypes.PLDM] += [mctp_resp]

    # Add responses to context
    ctx.mctp_responses = MctpResponseList(responses=responseList)
    return MctpResponseList(responses=responseList)



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum


class PldmTypeCodes(IntEnum):
    CONTROL = 0x00
    SMBIOS = 0x01
    PLATFORM_MONITORING = 0x02
    BIOS = 0x03
    FRU = 0x04
    FIRMWARE_UPDATES = 0x05
    RDE = 0x06
    OEM = 0x3F


class PldmControlCmdCodes(IntEnum):
    Reserved = 0x00
    SetTID = 0x01
    GetTID = 0x02
    GetPLDMVersion = 0x03
    GetPLDMTypes = 0x04
    GetPLDMCommands = 0x05
    SelectPLDMVersion = 0x06
    NegotiateTransferParameters = 0x07
    MultipartSend = 0x08
    MultipartReceive = 0x09


class CompletionCodes(IntEnum):
    SUCCESS = 0
    """The Request was accepted and completed normally."""

    ERROR = 1
    """This is a generic failure message. (It should not be used
    when a more specific result code applies.)"""

    ERROR_INVALID_DATA = 2
    """The packet payload contained invalid data or an illegal
    parameter value."""

    ERROR_INVALID_LENGTH = 3
    """The message length was invalid. (The Message body
    was larger or smaller than expected for the particular
    request.)"""

    ERROR_NOT_READY = 4
    """The Receiver is in a transient state where it is not ready
    to receive the corresponding message."""

    ERROR_UNSUPPORTED_CMD = 5
    """The command field in the control type of the received
    message is unspecified or not supported on this
    endpoint. This completion code shall be returned for any
    unsupported command values received in MCTP control
    Request messages."""

    ERROR_INVALID_PLDM_TYPE = 32
    """The PLDM Type field value in the PLDM request
    message is invalid or unsupported."""



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

"""
This module defines the ControlHdrPacket class and related enums and functions
for handling MCTP control header packets.
"""

from collections.abc import Callable
from enum import IntEnum
from typing import Any

from scapy.config import conf
from scapy.fields import (
    AnyField,
    BitEnumField,
    BitField,
    ByteEnumField,
    ConditionalField,
    Emph,
    Field,
    XBitField,
    XByteField,
)
from scapy.packet import Packet, bind_layers

from ...helpers import AllowRawSummary
from ...interfaces import ICanSetMySummaryClasses
from .. import EndpointContext
from ..transport import AutobindMessageType, MsgTypes, SmbusTransportPacket, TransportHdrPacket
from ..types import AnyPacketType
from . import ContrlCmdCodes, IControlMsgCanReply


class RqBit(IntEnum):
    RESPONSE = 0
    REQUEST = 1


@AutobindMessageType(MsgTypes.CTRL)
class ControlHdrPacket(AllowRawSummary, Packet):
    name = "MCTP-Control"
    fields_desc = [
        Emph(BitEnumField("rq", 0, 1, RqBit)),
        BitField("d", 0, 1),
        BitField("unused", 0, 1),
        XBitField("instance_id", 0, 5),
        Emph(ByteEnumField("cmd_code", 0, ContrlCmdCodes)),
        ConditionalField(XByteField("completion_code", 0), lambda pkt: pkt.rq == RqBit.RESPONSE.value),
    ]

    def do_dissect_payload(self, s: bytes) -> None:
        cls = self.guess_payload_class(s)
        try:
            p = cls(s, _internal=1, _underlayer=self)
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector and cls is not None:
                raise
            p = conf.raw_layer(s, _internal=1, _underlayer=self)
        self.add_payload(p)
        if isinstance(p, ICanSetMySummaryClasses):
            p.set_mysummary_classes([self.__class__, self.underlayer.__class__])

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        rqType = "REQ" if self.rq == RqBit.REQUEST.value else "RSP"
        summary = f"CONTROL {rqType} (instance_id: {self.instance_id}, cmd_code={self.cmd_code}"
        if self.rq == RqBit.RESPONSE.value:
            summary += f", completion_code={self.completion_code})"
        else:
            summary += ")"
        return summary, [TransportHdrPacket, SmbusTransportPacket]

    def answers(self, other: Packet) -> int:
        if not other.haslayer(ControlHdrPacket):
            return 0
        if self.rq != 0 or other.rq != 1:
            return 0
        if self.cmd_code != other.cmd_code:
            return 0
        if self.instance_id != other.instance_id:
            return 0
        return self.payload.answers(other.payload)

    def is_request(self, check_payload: bool = True) -> bool:
        return self.rq == 1

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        if not self.is_request():
            return None
        if self.cmd_code not in list(ContrlCmdCodes):
            return None

        completion_code = 0
        payload_resp = None
        if self.payload and isinstance(self.payload, IControlMsgCanReply):
            completion_code, payload_resp = self.payload.make_ctrl_reply(ctx)

        rsp = ControlHdr(
            rq=False,
            instance_id=self.instance_id,
            cmd_code=self.cmd_code,
            completion_code=completion_code,
        )
        return (rsp / payload_resp) if payload_resp else rsp


def ControlHdr(
    *args,
    rq: bool | RqBit = RqBit.RESPONSE,
    d: bool = False,
    instance_id: int = 0,
    cmd_code: ContrlCmdCodes = ContrlCmdCodes.GetEndpointID,
    completion_code: int | None = 0,
) -> ControlHdrPacket:
    if len(args):
        return ControlHdrPacket(*args)
    return ControlHdrPacket(
        rq=0 if not rq or rq in [False, RqBit.RESPONSE, 0] else 1,
        d=1 if d else 0,
        instance_id=instance_id,
        cmd_code=cmd_code,
        completion_code=completion_code,
    )


class AutobindControlMsg:
    def __init__(self, cmd_code: ContrlCmdCodes):
        self.cmd_code = cmd_code

    def __call__(self, cls: type[Packet]):
        cmd_code = self.cmd_code
        # print(f"Binding cls {cls} to cmd_code {cmd_code}:{self.is_request}")
        bind_layers(ControlHdrPacket, cls, cmd_code=cmd_code.value if type(cmd_code) == ContrlCmdCodes else cmd_code)
        if not hasattr(cls, "name") or cls.name is None:
            cls.name = cls.__name__
        if not hasattr(cls, "cmd_code") or cls.cmd_code is None:
            cls.cmd_code = self.cmd_code
        return cls


def set_control_fields(
    rq_fields: list[AnyField] | None = None, rsp_fields: list[AnyField] | None = None
) -> list[AnyField]:
    rq_fields = rq_fields or []
    rsp_fields = rsp_fields or []

    def gen_conditional_field(fld: AnyField, cond: Callable[[Packet], bool]):
        def default_cond(pkt):
            return False

        if isinstance(fld, ConditionalField):
            # unwrap the conditional field
            default_cond = fld.cond
            fld = fld.fld
        return ConditionalField(fld=fld, cond=lambda pkt: any([cond(pkt), default_cond(pkt)]))

    def is_request(pkt: Packet) -> bool:
        return pkt.underlayer.getfieldval("rq") == RqBit.REQUEST.value

    def is_response(pkt: Packet) -> bool:
        return pkt.underlayer.getfieldval("rq") == RqBit.RESPONSE.value

    fields = [gen_conditional_field(fld, is_request) for fld in rq_fields]
    fields += [gen_conditional_field(fld, is_response) for fld in rsp_fields]

    return fields


def response_fields(fields: list[AnyField]) -> list[AnyField]:
    """Wraps the list of Fields in a ConditionalField that checks if 'rq==0'"""
    return [ResponseField(fld) for fld in fields]


class ResponseField(ConditionalField):
    """Wraps the field within a ConditionalField that checks if 'rq==0'"""

    def __init__(self, fld: Field[Any, Any]) -> None:
        ConditionalField.__init__(self, fld, cond=lambda pkt: pkt.underlayer.getfieldval("rq") == 0)



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from scapy.fields import ByteField, FieldLenField, FieldListField
from scapy.packet import Packet

from .. import EndpointContext
from ..types import AnyPacketType
from .control import (
    AutobindControlMsg,
    ControlHdr,
    set_control_fields,
)
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


@AutobindControlMsg(ContrlCmdCodes.GetMessageTypeSupport)
class GetMessageTypeSupportPacket(Packet):
    name = "GetMessageTypeSupport"

    fields_desc = set_control_fields(
        rsp_fields=[
            FieldLenField("msg_type_cnt", None, fmt="!B", count_of="msg_type_list"),
            FieldListField("msg_type_list", [], ByteField("", 0), length_from=lambda pkt: pkt.msg_type_cnt),
        ]
    )

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        if not ctx.supported_msg_types:
            return CompletionCodes.ERROR_UNSUPPORTED_CMD, None
        # filter out the Control msg types
        msg_types = ctx.supported_msg_types
        # msg_types = [t for t in ctx.supported_msg_types if t != MsgTypes.MctpControl]
        return CompletionCodes.SUCCESS, GetMessageTypeSupportResponse(
            msg_types=msg_types,
        )


def GetMessageTypeSupport(_pkt: bytes | bytearray = b"", /, *args) -> GetMessageTypeSupportPacket:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.GetMessageTypeSupport)
    if _pkt:
        return GetMessageTypeSupportPacket(_pkt, _underlayer=hdr)
    return GetMessageTypeSupportPacket(_underlayer=hdr)


def GetMessageTypeSupportResponse(
    _pkt: bytes | bytearray = b"", /, *, msg_types: list[int] | None = None
) -> GetMessageTypeSupportPacket:
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.GetMessageTypeSupport)
    if _pkt:
        return GetMessageTypeSupportPacket(_pkt, _underlayer=hdr)
    return GetMessageTypeSupportPacket(
        msg_type_cnt=len(msg_types),
        msg_type_list=msg_types,
        # add a default underlayer to set the required "rq" field
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from .types import (
    CompletionCodes,
    PldmTypeCodes,
    PldmControlCmdCodes,
)

from .pldm import (
    RqBit,
    PldmHdr,
    PldmHdrPacket,
    AutobindPLDMMsg,
)

from .type1_base import (
    SetTIDPacket,
    GetTIDPacket,
    GetPLDMVersionPacket,
    GetPLDMTypesPacket,
    GetPLDMCommandsPacket,
)

from .type_2_platform_monitoring import (
    PldmPlatformMonitoringCmdCodes,
    PlatformEventMsgPacket,
    PlatformEventMsgClasses,
    PollForPlatformEventMsgPacket,
    GetSensorReadingPacket,
)



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from scapy.fields import ByteEnumField, MultipleTypeField, XByteField, XIntField, XShortField
from scapy.packet import Packet

from ...helpers import AllowRawSummary
from .. import EndpointContext
from ..types import AnyPacketType, VendorCapabilitySet, VendorIdFormat
from .control import (
    AutobindControlMsg,
    ControlHdr,
    set_control_fields,
)
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes

NO_MORE_CAPABILITY_SETS = 0xFF


@AutobindControlMsg(ContrlCmdCodes.GetVendorDefinedMessageSupport)
class GetVendorDefinedMessageSupportPacket(AllowRawSummary, Packet):
    fields_desc = set_control_fields(
        rq_fields=[
            XByteField("vendor_id_set_selector", 0),
        ],
        rsp_fields=[
            XByteField("next_vendor_id_set_selector", 0),
            ByteEnumField("vendor_id_format", 0, VendorIdFormat),
            MultipleTypeField(
                [
                    (XIntField("vendor_id", 0), lambda pkt: pkt.vendor_id_format == VendorIdFormat.IANA_ENT_NUMBER),
                ],
                XShortField("vendor_id", 0),
            ),
            XShortField("command_set_type", 0),
        ],
    )

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        if not ctx.supported_vdm_msg_types or self.vendor_id_set_selector >= len(ctx.supported_vdm_msg_types):
            # invalid request
            return CompletionCodes.ERROR_INVALID_DATA, None

        cap_set: VendorCapabilitySet = ctx.supported_vdm_msg_types[self.vendor_id_set_selector]

        next_set_selector = self.vendor_id_set_selector + 1
        if next_set_selector >= len(ctx.supported_vdm_msg_types):
            next_set_selector = NO_MORE_CAPABILITY_SETS

        return CompletionCodes.SUCCESS, GetVendorDefinedMessageSupportResponse(
            set_selector=next_set_selector,
            vendor_id_fmt=cap_set.id_format,
            vendor_id=cap_set.vendor_id,
            command_set_type=cap_set.command_set_type,
        )


def GetVendorDefinedMessageSupport(_pkt: bytes | bytearray = b"", /, *, set_selector: int = 0):
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.GetVendorDefinedMessageSupport)
    if _pkt:
        return GetVendorDefinedMessageSupportPacket(_pkt, _underlayer=hdr)
    return GetVendorDefinedMessageSupportPacket(
        vendor_id_set_selector=set_selector,
        _underlayer=hdr,
    )


def GetVendorDefinedMessageSupportResponse(
    _pkt: bytes | bytearray = b"",
    /,
    *,
    set_selector: int = 0,
    vendor_id_fmt: VendorIdFormat = VendorIdFormat.PCI_VENDOR_ID,
    vendor_id: int = 0,
    command_set_type: int = 0,
):
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.GetVendorDefinedMessageSupport)
    if _pkt:
        return GetVendorDefinedMessageSupportPacket(_pkt, _underlayer=hdr)
    return GetVendorDefinedMessageSupportPacket(
        next_vendor_id_set_selector=set_selector,
        vendor_id_format=vendor_id_fmt,
        vendor_id=vendor_id,
        command_set_type=command_set_type,
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from .aardvark_i2c import AardvarkI2CSocket
from .qemu_i2c_netdev import QemuI2CNetDevSocket
from .qemu_i3c_chardev import QemuI3CCharDevSocket



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum

from scapy.fields import XByteEnumField, XByteField
from scapy.packet import Packet

from .. import EndpointContext
from ..types import AnyPacketType
from .control import AutobindControlMsg, ControlHdr, set_control_fields
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


class AllocateEIDOperation(IntEnum):
    ALLOCATE_EIDS = 0
    FORCE_ALLOCATION = 1
    GET_ALLOCATION_INFO = 2


class AllocateEIDAllocationStatus(IntEnum):
    ACCEPTED = 0
    REJECTED = 1


@AutobindControlMsg(ContrlCmdCodes.AllocateEndpointIDs)
class AllocateEndpointIDsPacket(Packet):
    fields_desc = set_control_fields(
        rq_fields=[
            XByteEnumField("op", 0, AllocateEIDOperation),
            XByteField("allocated_pool_size", 0),
            XByteField("starting_eid", 0),
        ],
        rsp_fields=[
            XByteEnumField("status", 0, AllocateEIDAllocationStatus),
            XByteField("eid_pool_size", 0),
            XByteField("first_eid", 0),
        ],
    )

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        if not ctx.is_bus_owner:
            return CompletionCodes.ERROR_UNSUPPORTED_CMD, None

        status = AllocateEIDAllocationStatus.ACCEPTED
        eid_pool_size = ctx.pool_size if ctx.allocated_pool else 0
        first_eid = ctx.allocated_pool[0] if ctx.allocated_pool else 0

        if self.op != AllocateEIDOperation.GET_ALLOCATION_INFO:
            if self.op == AllocateEIDOperation.ALLOCATE_EIDS and ctx.allocated_pool:
                status = AllocateEIDAllocationStatus.REJECTED
            elif self.allocated_pool_size != 0:
                ctx.allocated_pool = list(range(self.starting_eid, self.starting_eid + self.allocated_pool_size))
                first_eid = ctx.allocated_pool[0]
            elif self.op == AllocateEIDOperation.FORCE_ALLOCATION:
                # Asking us to clear our EID allocation pool
                # TODO: signal to runtime that the routing table needs to be rebuilt
                ctx.allocated_pool = None
                first_eid = 0
        return CompletionCodes.SUCCESS, AllocateEndpointIDsResponse(
            status=status,
            eid_pool_size=eid_pool_size,
            first_eid=first_eid,
        )


def AllocateEndpointIDs(
    _pkt: bytes | bytearray = b"", /, *, op: AllocateEIDOperation, allocated_pool_size: int, starting_eid: int
) -> AllocateEndpointIDsPacket:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.AllocateEndpointIDs)
    if _pkt:
        return AllocateEndpointIDsPacket(_pkt, _underlayer=hdr)
    return AllocateEndpointIDsPacket(
        op=op,
        allocated_pool_size=allocated_pool_size,
        starting_eid=starting_eid,
        _underlayer=hdr,
    )


def AllocateEndpointIDsResponse(
    _pkt: bytes | bytearray = b"", /, *, status: AllocateEIDAllocationStatus, eid_pool_size: int, first_eid: int
):
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.AllocateEndpointIDs)
    if _pkt:
        return AllocateEndpointIDsPacket(_pkt, _underlayer=hdr)
    return AllocateEndpointIDsPacket(
        status=status,
        eid_pool_size=eid_pool_size,
        first_eid=first_eid,
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import crc8
from scapy.compat import raw
from scapy.config import conf
from scapy.fields import BitEnumField, BitField, ConditionalField, LenField, PacketLenField, XByteField
from scapy.layers.l2 import CookedLinux, CookedLinuxV2
from scapy.packet import Packet, bind_layers
from scapy.plist import PacketList

from ..helpers import AllowRawSummary, AnyPacketType
from ..interfaces import ICanSetMySummaryClasses, ICanVerifyIfRequest
from .types import EndpointContext, ICanReply, MsgTypes, Smbus7bitAddress


class TransportHdrPacket(AllowRawSummary, Packet):
    name = "MCTP-Transport"
    # match_subclass = True
    fields_desc = [
        BitField("rsvd", 0, 4),
        BitField("version", 0b0001, 4),
        XByteField("dst", 0),
        XByteField("src", 0),
        BitField("som", 0, 1),
        BitField("eom", 0, 1),
        BitField("pkt_seq", 0, 2),
        BitField("to", 0, 1),
        BitField("tag", 0, 3),
        # These are technically part of the msg, but are needed to properly bind the layers
        ConditionalField(BitEnumField("ic", 0, 1, {0: "no", 1: "yes"}), lambda pkt: pkt.som == 1),
        ConditionalField(BitEnumField("msg_type", MsgTypes.CTRL.value, 7, MsgTypes), lambda pkt: pkt.som == 1),
    ]

    def mysummary(self):  # type: () -> str
        summary = ""
        if self.underlayer and isinstance(self.underlayer, CookedLinux):
            summary += f"[{len(self.original):3}] / "
        summary += f"MCTP {self.pkt_seq}:{self.tag} ({self.dst:02X} <-- {self.src:02X}) ("
        flags = []
        msg_type = None
        if self.som == 1:
            flags += ["S"]
            msg_type = self.msg_type
        if self.eom == 1:
            flags += ["E"]
        if self.to == 1:
            flags += ["TO"]
        if flags:
            summary += ":".join(flags)
        summary += ")"
        # allow unknown message types to be passed in
        if msg_type is not None:
            msg_type_name = getattr(MsgTypes(msg_type), "name", f"{msg_type:02X}")
            summary += f" {msg_type_name}"
        return summary, [SmbusTransportPacket, TrimmedSmbusTransportPacket]

    def do_dissect_payload(self, s: bytes) -> None:
        if not s:
            return
        # Message Type is only present within the first fragment, just assume a raw payload
        cls = conf.raw_layer if self.som == 0 else self.guess_payload_class(s)
        try:
            p = cls(s, _internal=1, _underlayer=self)
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector and cls is not None:
                raise
            p = conf.raw_layer(s, _internal=1, _underlayer=self)
        self.add_payload(p)
        if isinstance(p, ICanSetMySummaryClasses):
            p.set_mysummary_classes([SmbusTransportPacket, TransportHdrPacket, TrimmedSmbusTransportPacket])

    def answers(self, rq_pkt: Packet) -> int:
        # the endpoint EID should match
        if rq_pkt.dst not in (self.src, 0):
            print(f"mismatched eid: {rq_pkt.dst} != {self.src}")
            return 0
        # the tags should be the same
        if self.tag != rq_pkt.tag:
            print(f"mismatched tags: {rq_pkt.tag} != {self.tag}")
            return 0
        # the request should be the tag owner
        if self.to == 0 and rq_pkt.to == 1:
            return self.payload.answers(rq_pkt.payload)
        print(f"mismatched to: {rq_pkt.to} != {self.to}")
        return 0

    def is_request(self, check_payload: bool = True) -> bool:
        # check next layer
        if check_payload and self.payload and isinstance(self.payload, ICanVerifyIfRequest):
            return self.payload.is_request()
        # check previous layer
        if self.underlayer and isinstance(self.underlayer, ICanVerifyIfRequest):
            return self.underlayer.is_request()
        # Fallback to checking the TO bit
        return self.to == 1

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        payload_resp = None
        if self.payload and isinstance(self.payload, ICanReply):
            payload_resp = self.payload.make_reply(ctx)
            if not payload_resp:
                return None

        if isinstance(payload_resp, PacketList | list):
            # TODO: Implement multiple payloads from upper layers
            msg = "multiple payloads are not yet supported"
            raise TypeError(msg)

        if not payload_resp:
            # TODO: should this just return None?
            return TransportHdr(
                msg_type=self.msg_type,
                to=False,
                tag=self.tag,
                src=ctx.assigned_eid or self.dst,
                dst=self.src,
                som=True,
                eom=True,
                pkt_seq=self.pkt_seq + 1,
            )

        # Split the response payload into packets (if needed)
        resps = PacketList()
        buf = raw(payload_resp)
        buf_size = msg_size = len(buf)
        buf_offset = 0
        pkt_seq = self.pkt_seq + 1
        while buf_size > 0:
            packet_size = min(ctx.mtu_size, buf_size)
            packet = payload_resp if msg_size <= ctx.mtu_size else buf[buf_offset : buf_offset + packet_size]
            som = bool(not buf_offset)
            eom = (buf_offset + packet_size) >= msg_size

            resp = TransportHdr(
                msg_type=self.msg_type,
                to=False,
                tag=self.tag,
                src=ctx.assigned_eid or self.dst,
                dst=self.src,
                som=som,
                eom=eom,
                pkt_seq=pkt_seq,
            )
            resps.append(resp / packet)

            buf_offset += packet_size
            buf_size -= packet_size
            pkt_seq += 1

        return resps


def TransportHdr(
    *args,
    dst: int = 0,
    src: int = 0,
    som: bool | int = True,
    eom: bool | int = True,
    pkt_seq: int = 0,
    to: bool | int = True,
    tag: int = 0,
    ic: bool | int = False,
    msg_type: MsgTypes = MsgTypes.CTRL,
    version: bytes = 0b0001,
) -> TransportHdrPacket:
    if len(args):
        return TransportHdrPacket(*args)
    return TransportHdrPacket(
        dst=dst,
        src=src,
        som=1 if som else 0,
        eom=1 if eom else 0,
        pkt_seq=pkt_seq,
        to=1 if to else 0,
        tag=tag,
        ic=1 if ic else 0,
        msg_type=msg_type,
        version=version,
    )


class ExtendedConditionalField(ConditionalField):
    """
    A FCS field that gets appended at the end of the *packet* (not layer).
    """

    __slots__ = ["fld", "cond"]

    def __init__(
        self,
        fld,  # type: Field[Any, Any]
        cond,  # type: Callable[[Packet, bytes], bool]
    ):
        # type: (...) -> None
        self.fld = fld
        self.cond = cond

    def _evalcond(self, pkt, s=None):
        # type: (Packet, bytes) -> bool
        return bool(self.cond(pkt, s or pkt.original))

    def i2h(self, pkt, val):
        # type: (Optional[Packet], Any) -> Any
        if pkt and not self._evalcond(pkt, pkt.original or [val]):
            return None
        return self.fld.i2h(pkt, val)

    def getfield(self, pkt, s):
        # type: (Packet, bytes) -> Tuple[bytes, Any]
        if self._evalcond(pkt, s):
            return self.fld.getfield(pkt, s)
        return s, None

    def addfield(self, pkt, s, val):
        # type: (Packet, bytes, Any) -> bytes
        if self._evalcond(pkt, s):
            return self.fld.addfield(pkt, s, val)
        return s


class SmbusTransportPacket(AllowRawSummary, Packet):
    name = "SMBUS/I2C"

    fields_desc = [
        ExtendedConditionalField(
            XByteField("dst_addr", 0), lambda pkt, s: (s[0] != 0x0F) if s else pkt.getfieldval("dst_addr")
        ),
        XByteField("command_code", 0x0F),
        LenField("byte_count", None, fmt="B", adjust=lambda x: 0 if not x else (x + 1)),
        XByteField("src_addr", 0),
        PacketLenField("load", None, TransportHdrPacket, length_from=lambda x: x.byte_count - 1),
        ExtendedConditionalField(
            XByteField("pec", None), lambda pkt, s: len(s) == 1 or len(s) >= pkt.getfieldval("byte_count")
        ),
    ]

    def mysummary(self):  # type: () -> str
        summary = "SMBUS ("
        if "dst_addr" in self.fields:
            summary += f"dst=0x{self.dst_addr:02X}, "
        summary += f"src=0x{self.src_addr:02X}, byte_count={self.byte_count}"
        if "pec" in self.fields:
            summary += f", pec=0x{self.pec:02X}"
        summary += ")"
        return summary, [SmbusTransportPacket]

    def post_build(self, p, pay):
        # hexdump(p)
        # hexdump(pay)
        p += pay
        if self.pec is None:
            crc = crc8.crc8()
            crc.update(p[:-1])
            val = crc.digest()
            self.pec = int.from_bytes(val, byteorder="little")
            p = p[:-1] + val
        elif pay and self.pec != pay[-1]:
            p = p[:-1] + int.to_bytes(self.pec, byteorder="little", length=1)
        return p

    def answers(self, other: Packet) -> int:
        if self.src_addr >> 1 != other.dst_addr >> 1:
            return 0
        if self.command_code != other.command_code:
            return 0
        if self.payload:
            return self.payload.answers(other.payload)
        if self.load:
            return self.load.answers(other.load)
        return None

    def src_addr_7bit(self) -> Smbus7bitAddress:
        return Smbus7bitAddress(self.src_addr >> 1)

    def dst_addr_7bit(self) -> Smbus7bitAddress:
        return Smbus7bitAddress(self.dst_addr >> 1)

    def is_request(self, check_payload: bool = True) -> bool:
        if check_payload and self.payload and isinstance(self.payload, ICanReply):
            return self.payload.is_request()
        if check_payload and self.load and isinstance(self.load, ICanReply):
            return self.load.is_request()
        return False

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        # Validate the smbus address (if set in context)
        if ctx.physical_address:
            if not isinstance(ctx.physical_address, Smbus7bitAddress):
                return None
            if ctx.physical_address != self.dst_addr_7bit():
                print(
                    f"Warning: received incorrectly addressed packet: {self.dst_addr_7bit()} != {ctx.physical_address}"
                )
                return None

        payload_resp = None
        if self.payload and isinstance(self.payload, ICanReply):
            payload_resp = self.payload.make_reply(ctx)
            if not payload_resp:
                return None
        if self.load and isinstance(self.load, ICanReply):
            payload_resp = self.load.make_reply(ctx)
            if not payload_resp:
                return None

        dst = self.dst_addr_7bit()
        src = self.src_addr_7bit()

        if not isinstance(payload_resp, PacketList | list):
            payload_resp = [payload_resp]

        # wrap each packet with the SMBUS transport header
        packets = PacketList()
        for packet in payload_resp:
            resp_hdr = SmbusTransport(
                dst_addr=src.write(),
                src_addr=dst.read(),
                byte_count=len(packet),
                load=packet,
            )

            packets.extend(resp_hdr)

        return packets


class TrimmedSmbusTransportPacket(SmbusTransportPacket):
    name = "SMBUS/I2C"

    fields_desc = [
        XByteField("command_code", 0x0F),
        LenField("byte_count", None, fmt="B", adjust=lambda x: 0 if not x else (x + 1)),
        XByteField("src_addr", 0),
        PacketLenField("load", None, TransportHdrPacket, length_from=lambda x: x.byte_count - 1),
        ExtendedConditionalField(XByteField("pec", None), lambda pkt, s: len(s) == 1 or len(s) >= pkt.byte_count),
    ]


def SmbusTransport(
    *args,
    dst_addr: int | Smbus7bitAddress = 0,
    src_addr: int | Smbus7bitAddress = 1,
    byte_count: int | None = None,
    command_code: int = 0x0F,
    load: AnyPacketType = None,
    pec: int | None = None,
) -> SmbusTransportPacket:
    if len(args):
        return SmbusTransportPacket(*args)
    if isinstance(dst_addr, Smbus7bitAddress):
        dst_addr = dst_addr.write()
    if isinstance(src_addr, Smbus7bitAddress):
        src_addr = src_addr.read()
    if not byte_count:
        byte_count = len(load) if load else 0
    byte_count += 1
    if not pec:
        crc = crc8.crc8()
        crc.update(bytes([dst_addr, command_code, byte_count, src_addr]))
        crc.update(raw(load))
        val = crc.digest()
        pec = int.from_bytes(val, byteorder="little")
    return SmbusTransportPacket(
        dst_addr=dst_addr, src_addr=src_addr, command_code=command_code, byte_count=byte_count, load=load, pec=pec
    )


class AutobindMessageType:
    def __init__(self, msg_type: MsgTypes):
        self.msg_type = msg_type

    def __call__(self, cls: type[Packet]):
        # print(f"Binding cls {cls} to msg_type {self.msg_type}")
        bind_layers(TransportHdrPacket, cls, msg_type=self.msg_type.value)
        if not hasattr(cls, "name") or cls.name is None:
            cls.name = cls.__name__
        return cls


# Add the MCTP Transport (ETH_P_MCTP = 0xFA) as a valid L2 protocol
bind_layers(CookedLinux, TransportHdrPacket, proto=0xFA)
bind_layers(CookedLinuxV2, TransportHdrPacket, proto=0xFA)

# Add the MCTP Transport as a valid SMBUS command protocol
bind_layers(SmbusTransportPacket, TransportHdrPacket, command_code=0x0F)



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from .types import (
    EndpointContext,
    MsgTypes,
    Smbus7bitAddress,
    Smbus10bitAddress,
    VendorCapabilitySet,
    VendorIdFormat,
    AnyPhysicalAddress,
)

from .transport import (
    TransportHdr,
    TransportHdrPacket,
    SmbusTransport,
    SmbusTransportPacket,
    TrimmedSmbusTransportPacket,
)

# Import the main Packets for each MsgType to perform autobinding
from .control import ControlHdrPacket
from .pldm import PldmHdrPacket
from .vdpci import VdPciHdrPacket

# Import any utilities
from .context_utils import import_pcap_dump



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import dataclasses
import pickle
import threading
from dataclasses import field
from enum import Enum
from threading import Thread
from typing import Any, Protocol, runtime_checkable

from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import BaseConfig
from scapy.supersocket import SuperSocket

from pymctp.automaton import EndpointSession, SimpleEndpointAM
from pymctp.exerciser import AardvarkI2CSocket, QemuI2CNetDevSocket, QemuI3CCharDevSocket
from pymctp.layers.mctp import EndpointContext, Smbus7bitAddress


class ConfigTypes(str, Enum):
    Socket = "socket"
    Aardvark = "aardvark"
    CharDev = "chardev"


@runtime_checkable
class ISessionConfig(Protocol):
    socket: SuperSocket

    def close_socket(self):
        pass


@dataclasses.dataclass()
class CharDevSocketConfig(DataClassDictMixin):
    type = ConfigTypes.CharDev
    in_file: str
    name: str
    pid: int
    bcr: int
    dcr: int
    mwl: int = 256
    mrl: int = 256
    dynamic_addr: int = 0

    socket: QemuI3CCharDevSocket | None = field(
        default=None, init=False, metadata={"serialize": pickle.dumps, "deserialize": pickle.loads}
    )

    def __post_init__(self):
        self.socket = QemuI3CCharDevSocket(
            in_file=self.in_file,
            id_str=self.name,
            pid=self.pid,
            bcr=self.bcr,
            dcr=self.dcr,
            mwl=self.mwl,
            mrl=self.mrl,
            dynamic_addr=self.dynamic_addr,
        )

    def close_socket(self):
        self.socket.close()


@dataclasses.dataclass()
class UdpSocketConfig(DataClassDictMixin):
    type = ConfigTypes.Socket
    in_port: int
    out_port: int
    name: str
    iface: str = "localhost"
    dump_hex: bool = True
    dump_packet: bool = False

    socket: QemuI2CNetDevSocket | None = field(
        default=None, init=False, metadata={"serialize": pickle.dumps, "deserialize": pickle.loads}
    )

    def __post_init__(self):
        self.socket = QemuI2CNetDevSocket(
            iface=self.iface,
            in_port=self.in_port,
            out_port=self.out_port,
            id_str=self.name,
            dump_hex=self.dump_hex,
            dump_packet=self.dump_packet,
        )

    def close_socket(self):
        self.socket.close()


def deserialize_aardvark_address(value: str | int | Smbus7bitAddress) -> Smbus7bitAddress:
    if isinstance(value, Smbus7bitAddress):
        return value
    return Smbus7bitAddress(address=int(value))


@dataclasses.dataclass()
class AardvarkConfig(DataClassDictMixin):
    type = ConfigTypes.Aardvark
    slave_addr: Smbus7bitAddress = field(metadata=field_options(alias="slave_address"))
    serial_number: str
    name: str
    dump_hex: bool = True
    dump_packet: bool = False

    socket: AardvarkI2CSocket | None = field(
        default=None, init=False, metadata={"serialize": pickle.dumps, "deserialize": pickle.loads}
    )

    class Config(BaseConfig):
        serialization_strategy = {Smbus7bitAddress: {"deserialize": deserialize_aardvark_address}}

    def __post_init__(self):
        self.socket = AardvarkI2CSocket(
            slave_address=self.slave_addr,
            serial_number=self.serial_number,
            id_str=self.name,
            dump_hex=self.dump_hex,
            dump_packet=self.dump_packet,
        )

    def create_session(self) -> EndpointSession:
        pass

    def close_socket(self):
        self.socket.close()


def deserialize_supersocket(value: dict) -> AardvarkConfig | UdpSocketConfig | CharDevSocketConfig:
    config_type = value.get("type")
    if config_type == ConfigTypes.Socket:
        return UdpSocketConfig.from_dict(value)
    if config_type == ConfigTypes.Aardvark:
        return AardvarkConfig.from_dict(value)
    if config_type == ConfigTypes.CharDev:
        return CharDevSocketConfig.from_dict(value)
    msg = f"Unknown config type {config_type}"
    raise ValueError(msg)


@dataclasses.dataclass()
class EndpointConfig(DataClassDictMixin):
    context: EndpointContext
    config: AardvarkConfig | UdpSocketConfig | CharDevSocketConfig
    thread_kwargs: dict[str, Any] = field(default_factory=dict)

    class Config(BaseConfig):
        serialization_strategy = {
            AardvarkConfig | UdpSocketConfig | CharDevSocketConfig: {"deserialize": deserialize_supersocket}
        }


@dataclasses.dataclass()
class EndpointManager:
    config: EndpointConfig
    session: EndpointSession
    socket: SuperSocket
    thread: Thread
    am: SimpleEndpointAM

    @classmethod
    def from_config(cls, config: dict[Any, Any], start_thread=True, verbose: bool = False):
        cfg = EndpointConfig.from_dict(config)
        print(f"DEBUG: {cfg or 'None'}")
        socket = cfg.config.socket
        session = EndpointSession(context=cfg.context, socket=socket)

        am = SimpleEndpointAM(socket=socket, context=cfg.context, session=session, verbose=verbose)
        if cfg.context.is_bus_owner:
            # TODO: add discovery flow answering machine here
            pass
        thread = threading.Thread(target=am, kwargs=cfg.thread_kwargs)
        if start_thread:
            thread.start()
        return EndpointManager(
            socket=socket,
            config=cfg,
            session=session,
            thread=thread,
            am=am,
        )

    @property
    def context(self) -> EndpointContext:
        return self.config.context

    @property
    def supersocket(self) -> SuperSocket:
        return self.config.config.socket

    def stop_sniffer(self):
        if self.am.sniffer.running:
            self.am.stop()



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from scapy.config import conf
from scapy.packet import Packet

from .interfaces import (
    AnyPacketType,
    ICanSetMySummaryClasses,
    ICanVerifyIfRequest,
)


class AllowRawSummary:
    def do_dissect_payload(self, s: bytes) -> None:
        if not s:
            return
        cls = self.guess_payload_class(s)
        try:
            p = cls(s, _internal=1, _underlayer=self)
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector and cls is not None:
                raise
            p = conf.raw_layer(s, _internal=1, _underlayer=self)
        self.add_payload(p)
        if isinstance(p, ICanSetMySummaryClasses):
            p.set_mysummary_classes([self.__class__, self.underlayer.__class__])

    def _do_summary(self):
        # type: () -> Tuple[int, str, List[Any]]
        if "load" in self.fields:
            found, s, needed = self.load._do_summary()  # noqa: SLF001
            if self.payload and self.payload.original:
                pld = conf.raw_layer(self.payload.original)
                s = f"{s} / {pld}"
        else:
            found, s, needed = self.payload._do_summary()  # noqa: SLF001
        ret = ""
        # if not found or self.__class__ in needed:
        ret = self.mysummary()
        if isinstance(ret, tuple):
            ret, n = ret
            needed += n
        if ret or needed:
            found = 1
        if not ret:
            ret = self.__class__.__name__ if self.show_summary else ""
        if self.__class__ in conf.emph:
            impf = [f"{f.name}={f.i2repr(self, self.getfieldval(f.name))}" for f in self.fields_desc if f in conf.emph]
            ret = f"{ret} [{' '.join(impf)}]"
        ret = f"{ret} / {s}" if ret and s else f"{ret}{s}"
        return found, ret, needed



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import threading
from collections.abc import Callable
from typing import TYPE_CHECKING, NamedTuple

from scapy.packet import Packet
from scapy.sendrecv import sndrcv
from scapy.sessions import DefaultSession
from scapy.supersocket import SuperSocket

from ..layers.mctp import (
    AnyPhysicalAddress,
    EndpointContext,
    Smbus7bitAddress,
    SmbusTransport,
    TransportHdr,
    TransportHdrPacket,
)
from ..layers.mctp.control import ControlHdr, ControlHdrPacket, IControlMsgPacket
from ..layers.mctp.types import AnyPacketType, MsgTypes

if TYPE_CHECKING:
    from scapy.ansmachine import AnsweringMachine
    from scapy.plist import PacketList


class HandlerResponse(NamedTuple):
    stop_processing: bool
    reply: AnyPacketType


class EndpointSession(DefaultSession):
    """Manages sending and receiving MCTP messages on the specified socket for a single endpoint."""

    def __init__(self, *args, context: EndpointContext, socket: SuperSocket, **kwargs):
        """
        Overloaded the DefaultSession.__init__() to add type hints for class specific parameters.
        :param args: Arguments to pass to DefaultSession
        :param context: MCTP Endpoint content to store runtime state
        :param socket: Socket to send and receive packets
        :param kwargs: Any additional arguments available in Scapy DefaultSession
        """
        super().__init__(*args, **kwargs)
        self.context = context
        self.socket = socket
        self.pending_rqs: list[tuple[Packet, Callable[[Packet], None]]] = []
        self._next_tag_number = 0
        self._next_instance_id = 0
        self._responder_lock = threading.Lock()

        # Not allowed during init due to the AnsweringMachine needing the session to initialize properly.
        # Instead, the AnsweringMachine will register itself during initialization.
        self.am: AnsweringMachine | None = None
        self.default_handlers: dict[type, Callable[[Packet, EndpointContext], HandlerResponse | None]] = {}

    def register_handler(self, layer_cls: type, handler: Callable[[Packet, EndpointContext], HandlerResponse | None]):
        """
        Registers the specified handler to allow the handler to be invoked anytime a msg is received that
        contains the specified layer class.
        :param layer_cls: Class that needs to be present to trigger handler (should be the lowest layer possible)
        :param handler: Callback handler when matching msg is received
        """
        self.default_handlers[layer_cls] = handler

    @property
    def next_tag_number(self) -> int:
        """
        Generates a tag number that automatically increments (and rolls over) on each access
        :return: the next tag number generated
        """
        next_tag = self._next_tag_number
        self._next_tag_number = (next_tag + 1) % 8
        return next_tag

    @property
    def next_instance_id(self) -> int:
        """
        Generates an instance id that automatically increments (and rolls over) on each access

        :return: the next instance id number generated
        """
        next_id = self._next_instance_id
        self._next_instance_id = (next_id + 1) % 32
        return next_id

    def on_packet_received(self, rq_pkt: Packet | None) -> None:
        """
        Will be called by sniff() for each received packet (that passes the filters).

        :param rq_pkt: The received packet
        :return: None
        """
        if not rq_pkt or not rq_pkt.haslayer(TransportHdrPacket):
            return

        mctp_pkt_hdr = rq_pkt.getlayer(TransportHdrPacket)
        is_request = bool(self.am.is_request(rq_pkt))
        som = bool(mctp_pkt_hdr.som)
        eom = bool(mctp_pkt_hdr.eom)

        if som and not eom:
            # TODO: handle packet fragments (see ISOTPSession for an example)
            msg = "TODO: handle receiving fragemented packets"
            raise RuntimeError(msg)

        # Lock the session to prevent another thread trying to send a request on the bus before we finish replying
        with self._responder_lock:
            plist = self.pending_rqs

            # 1) look for responses to pending requests
            for i, pending_rq in enumerate(plist):
                sentpkt, callback = pending_rq
                if not rq_pkt.answers(sentpkt):
                    continue
                del self.pending_rqs[i]
                callback(rq_pkt)
                return

            # 2) Check if we have a handler registered for this type of packet
            stop_processing = False
            response_pkts: PacketList = []
            for cls, handler in self.default_handlers.items():
                if cls in rq_pkt:
                    resp: HandlerResponse = handler(rq_pkt, self.context)
                    if not resp:
                        resp = HandlerResponse(False, None)
                    stop_processing = resp.stop_processing
                    if resp.reply:
                        response_pkts.append(resp.reply)
                        if stop_processing:
                            break
                    elif stop_processing:
                        return

            # 3) if a request, use the default reply methods to generate response
            if is_request and not stop_processing:
                reply = self.am.make_reply(rq_pkt)
                if reply:
                    response_pkts += reply

            # send any queued replies
            if response_pkts:
                self.am.send_reply(response_pkts)

    def sndrcv_control_msg(
        self,
        pkt: Packet,
        dst_eid: int,
        dst_phy_addr: AnyPhysicalAddress,
        *,
        instance_id: int | None = None,
        timeout_s: float | None = None,
        threaded: bool = False,
    ) -> Packet | None:
        """
        Sends the specified MCTP Control Payload to the destination and waits for a
        response. This method will detect if the ControlHdrPacket, TransportHdrPacket,
        and the SmbusTransportPacket are present and add them if missing.

        Example:

            >>> rsp = sndrcv_control_msg(GetEndpointID(), 0x15, Smbus7bitAddress(0x20 >> 1))

        :param pkt: The MCTP Control Msg packet to send
        :param dst_eid: The destination endpoint id
        :param dst_phy_addr: The physical address of the destination (endpoint or bridge/next hop)
        :param instance_id: The instance to differentiate different MCTP Control requests (autogenerated if not present)
        :param timeout_s: Max time to wait for a response (in seconds)
        :param threaded: Set to `True` if
        :return: the received packet or None
        """
        if not isinstance(pkt, IControlMsgPacket):
            msg = "Packet is malformed, does not implement the IControlMsgPacket protocol"
            raise TypeError(msg)
        cmd_code = pkt.cmd_code

        if not pkt.haslayer(ControlHdrPacket):
            pkt = (
                ControlHdr(
                    rq=True,
                    cmd_code=cmd_code,
                    instance_id=(instance_id or self.next_instance_id),
                )
                / pkt
            )
        return self.sndrcv_mctp_msg(
            pkt=pkt, dst_eid=dst_eid, dst_phy_addr=dst_phy_addr, timeout_s=timeout_s, threaded=threaded
        )

    def sndrcv_mctp_msg(
        self,
        pkt: Packet,
        dst_eid: int,
        dst_phy_addr: AnyPhysicalAddress,
        *,
        msg_type: MsgTypes = MsgTypes.CTRL,
        msg_tag: int | None = None,
        timeout_s: float | None = None,
        threaded: bool = False,
    ) -> Packet | None:
        # TODO: handle packet fragments
        pkt = (
            TransportHdr(
                src=self.context.eid,
                dst=dst_eid,
                som=1,
                eom=1,
                pkt_seq=0,
                to=1,
                tag=msg_tag or self.next_tag_number,
                msg_type=msg_type,
            )
            / pkt
        )

        src_phy_addr = self.context.physical_address
        if isinstance(dst_phy_addr, Smbus7bitAddress) and isinstance(src_phy_addr, Smbus7bitAddress):
            pkt = SmbusTransport(dst_addr=dst_phy_addr, src_addr=src_phy_addr) / pkt
        else:
            msg = f"Only Smbus7bitAddress are supported: {type(dst_phy_addr)}"
            raise TypeError(msg)

        # hexdump(pkt)
        # if drop_pec:
        #     data = raw(pkt)
        #     pkt = Raw(data[:-1])

        if threaded or (self.am and hasattr(self.am, "sniffer") and self.am.sniffer.running):
            return self.sr1threaded(pkt=pkt, timeout_s=timeout_s)
        return self.sr1(pkt=pkt, timeout_s=timeout_s)

    def sr1(self, pkt: Packet, timeout_s: float | None = None) -> Packet | None:
        """
        Sends and receives one packet on the socket. This method will stall
        the main thread until a response is received or the timeout is reached.

        :param pkt: The packet to send
        :param timeout_s: Max time to wait for a response (in seconds)
        :return: the received packet or None
        """
        with self._responder_lock:
            a, b = sndrcv(self.socket, pkt, timeout=int(timeout_s) if timeout_s else None, session=self)
            if len(a) > 0:
                return a[0][1]
        return None

    def sr1threaded(self, pkt: Packet, timeout_s: float | None = None) -> Packet | None:
        """
        Sends and receives on packet on the socket. This method will queue the response
        handler (callback) and start a timer to handle no response. This method is meant
        to be used when an AnsweringMachine is already running in the session (as there
        can only be one service bound to the socket at a time). The AnsweringMachine will
        invoke the "on_packet_received()" method when a response is received, which will
        check if the response is for a pending request.

        :param pkt: The packet to send
        :param timeout_s: Max time to wait for a response (in seconds)
        :return: The received packet or None
        """
        completed = threading.Event()
        rsp: list[Packet] = []

        # print(f"Sending request using threading")

        def callback(pkt: Packet):
            # print(f"response received, triggering callback")
            rsp.append(pkt)
            completed.set()

        # queue event
        item = (pkt, callback)
        self.pending_rqs.append(item)

        # send event
        with self._responder_lock:
            self.socket.send(pkt)

        # wait for completion
        timed_out = completed.wait(timeout=timeout_s)
        if not timed_out or not rsp:
            # remove the item from the list (ignoring any errors)
            try:
                with self._responder_lock:
                    self.pending_rqs.remove(item)
            finally:
                pass
            return None
        return rsp[0]



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import random
import time
from collections.abc import Callable
from typing import cast

from scapy.ansmachine import AnsweringMachine
from scapy.packet import Packet
from scapy.plist import PacketList, _PacketIterable  # imported for type hinting on overloaded methods
from scapy.sendrecv import AsyncSniffer
from scapy.supersocket import SuperSocket

from ..layers.interfaces import AnyPacketType
from ..layers.mctp.control import ControlHdrPacket
from ..layers.mctp.types import EndpointContext, ICanReply
from .sessions import EndpointSession


class SimpleEndpointAM(AnsweringMachine):
    """AnsweringMachine which emulates the basic behavior of a real MCTP endpoint.

    Usage:
        >>> import threading
        >>> from pymctp.automaton import SimpleEndpointAM
        >>> from pymctp.layers.mctp import *
        >>> from pymctp.exerciser import QemuI2CNetDevSocket
        >>> context = EndpointContext(physical_address=Smbus7bitAddress(0x20 >> 1), \
                              supported_msg_types=[MsgTypes.CTRL, MsgTypes.PLDM])
        >>> sock = QemuI2CNetDevSocket(iface="127.0.0.1", in_port=5559, out_port=5558)
        >>> session = EndpointSession(context=context, socket=sock)
        >>> am = SimpleEndpointAM(socket=sock, context=context, session=session)
        >>> sim = threading.Thread(target=am, kwargs={'count': 10, 'timeout':5*60})
        >>> sim.start()
        >>> rsp = session.sndrcv_control_msg(GetEndpointID(), 0x15, Smbus7bitAddress(0x20 >> 1))
        >>> rsp.show2() if rsp else print(f"No response received, timeout?")
    """

    context: EndpointContext
    socket: SuperSocket

    function_name = "simple_bus_owner"
    # removed the "iface", "promisc", "count", and "type" options as they are ethernet specific
    sniff_options_list = [
        "store",
        "opened_socket",
        "count",
        "filter",
        "prn",
        "stop_filter",
        "timeout",  # added timeout to stop sniffer after designated time period
    ]
    # removed the "socket" option to allow it to be passed to "parse_options"
    send_options_list = ["iface", "inter", "loop", "verbose"]

    def __init__(
        self,
        session: EndpointSession | None = None,
        socket: SuperSocket | None = None,
        context: EndpointContext | None = None,
        timeout: float | None = None,
        **kwargs,
    ):
        """
        Overloaded the AnsweringMachine.__init__() to add type hints for class specific parameters.

        :param session: Endpoint Session which oversees AM interactions
        :param socket: Socket to send and receive packets
        :param context: MCTP Endpoint content to store runtime state
        :param timeout: Specifies the sniffing timeout (seconds)
        :param kwargs: Any additional sniff/send options available in Scapy AnsweringMachine
        """
        self.context = context or EndpointContext()
        self.socket = socket
        self.session = session
        if self.session:
            self.session.am = self

        self.sniffer: AsyncSniffer | None = None
        super().__init__(timeout=timeout, **kwargs)

    def sniff(self) -> PacketList | None:
        """
        Overloaded the AnsweringMachine.sniff() method to always capture the sniffer in an instance attribute
        """
        self.sniffer = AsyncSniffer()
        self.sniffer._run(**self.optsniff)  # noqa: SLF001
        return cast(PacketList, self.sniffer.results)

    def parse_options(
        self,
        session: EndpointSession | None = None,
        socket: SuperSocket | None = None,
        timeout: float | None = None,
    ) -> None:
        """
        Sets up any additional sniffing/sending options that are not part of the
        standard AnsweringMachine.

        :param session: Endpoint Session which oversees AM interactions
        :param socket: Socket to send and receive packets
        :param timeout: Specifies the sniffing timeout (seconds)
        """
        if self.session:
            self.sniff_options["session"] = self.session
        self.sniff_options["timeout"] = timeout
        self.sniff_options["opened_socket"] = self.socket

    def is_request(self, req: Packet) -> int:
        """
        Called within AnsweringMachine.reply() to determine if the received
        packet requires a response.
        :param req: The received packet
        :return: 0 if no reply is necessary and 1 if a reply is required.
        """
        if isinstance(req, ICanReply):
            return req.is_request()
        return req.haslayer(ControlHdrPacket) and req.getlayer(ControlHdrPacket).rq == 1

    def make_reply(self, req: Packet | ICanReply) -> _PacketIterable:
        """
        Creates a reply to the incoming request (pre-confirmed by is_request())

        :param req: The received request packet
        :return: The fully formed response packet, or None (if no response can be generated)
        """
        rsp = None
        if isinstance(req, ICanReply):
            rsp = req.make_reply(self.context)
        # TODO: add any custom responses here
        return rsp

    def send_reply(self, reply: _PacketIterable, send_function: Callable[..., None] | None = None) -> None:
        """
        Sends the reply packets (sequentially) on the socket or using the "send_function" (if specified).

        @note This method does not wait for a response before sending the next reply

        :param reply: The fully formed packet(s) to send
        :param send_function: An optional callable method to invoke to send each packet (instead of the socket)
        """
        for p in reply:
            if len(reply) > 1:
                time.sleep(random.uniform(0.001, 0.01))  # noqa: S311
            else:
                time.sleep(random.uniform(0.1, 0.25))  # noqa: S311
            if self.socket:
                self.socket.send(p)

    def print_reply(self, req: AnyPacketType, reply: AnyPacketType) -> None:
        if isinstance(reply, PacketList):
            print("")
            print(f"{req.summary()} ==> {[res.summary() for res in reply]}")
        else:
            print(f"{req.summary()} ==> {reply.summary()}")

    @property
    def sniffer_running(self):
        return self.sniffer.running

    def stop_sniffer(self, join: bool = False) -> PacketList | None:
        if self.sniffer_running:
            return self.sniffer.stop(join=join)
        return None



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import time
from enum import IntEnum

from scapy.config import conf
from scapy.fields import BitEnumField, BitField, ShortEnumField, XByteField
from scapy.packet import Packet, Raw, bind_layers

from ...interfaces import ICanVerifyIfRequest
from .. import EndpointContext
from ..transport import (
    AutobindMessageType,
    MsgTypes,
    SmbusTransportPacket,
    TransportHdrPacket,
    TrimmedSmbusTransportPacket,
)
from ..types import AnyPacketType
from ...interfaces import ICanSetMySummaryClasses
from .types import VdPCIVendorIds


class RqBit(IntEnum):
    RESPONSE = 0
    REQUEST = 1


DEFAULT_HDR_VERSION = 0


@AutobindMessageType(MsgTypes.VDPCI)
class VdPciHdrPacket(Packet):
    name = "VDM-PCI"
    fields_desc = [
        ShortEnumField("vendor_id", 0, VdPCIVendorIds),
        BitEnumField("rq", 0, 1, RqBit),
        BitField("rsv", 0, 2),
        BitField("unused", 0, 5),
        XByteField("vdm_cmd_code", 0),
    ]

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        rqType = "REQ" if self.is_request() else "RSP"
        summary = (
            f"{self.name} {rqType} (VID: {self.vendor_id:04X}, cmd_code: 0x{self.vdm_cmd_code:02X}, rq: {self.rq})"
        )
        return summary, [TransportHdrPacket, SmbusTransportPacket, TrimmedSmbusTransportPacket]

    def do_dissect_payload(self, s: bytes) -> None:
        if not s:
            return
        cls = self.guess_payload_class(s)
        try:
            p = cls(s, _internal=1, _underlayer=self)
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector and cls is not None:
                raise
            p = conf.raw_layer(s, _internal=1, _underlayer=self)
        self.add_payload(p)
        if isinstance(p, ICanSetMySummaryClasses):
            p.set_mysummary_classes([VdPciHdrPacket, TransportHdrPacket])

    def answers(self, other: Packet) -> int:
        if self.rq != 0 or other.rq != 1:
            return 0
        if self.vendor_id != other.vendor_id:
            return 0
        if self.vdm_cmd_code != other.vdm_cmd_code:
            return 0
        return self.payload.answers(other.payload)

    def is_request(self, check_payload: bool = True) -> bool:
        return any(
            [
                self.rq == RqBit.REQUEST.value,
                self.payload and isinstance(self.payload, ICanVerifyIfRequest) and self.payload.is_request(),
                self.underlayer
                and isinstance(self.underlayer, ICanVerifyIfRequest)
                and self.underlayer.is_request(check_payload=False),
            ]
        )

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        if not self.is_request():
            return None
        if self.vendor_id not in list(VdPCIVendorIds):
            return None
        # only make a reply if we are a supported msg type (allows context to control which packets generate responses)
        if MsgTypes.VDPCI not in ctx.supported_msg_types:
            return None

        payload_resp = None

        # TODO: fill in reading from the file
        if ctx.mctp_responses:
            vdpci_hdr: VdPciHdrPacket = self.getlayer(VdPciHdrPacket)
            hdr_data = bytes(vdpci_hdr)
            data = bytes(vdpci_hdr.payload)
            vendor_id = self.vendor_id
            resp_info = ctx.get_response(MsgTypes.VDPCI, data, vendor_id.name, str(self.vdm_cmd_code))
            if resp_info:
                print(f"***> VDPCI Request Matched: {resp_info.description}")
                resp_data = Raw(resp_info.data)
                delay = resp_info.processing_delay
                if delay:
                    time.sleep(delay / 1000.0)
                return Raw(bytes([hdr_data[0], hdr_data[1], 0, hdr_data[3]])) / resp_data

        rsp = VdPciHdr(
            rq=False,
            vendor_id=self.vendor_id,
            vdm_cmd_code=self.vdm_cmd_code,
        )
        return (rsp / payload_resp) if payload_resp else rsp


def VdPciHdr(*args, rq: bool | RqBit = RqBit.RESPONSE, vendor_id: int = 0, vdm_cmd_code: int = 0) -> VdPciHdrPacket:
    if len(args):
        return VdPciHdrPacket(*args)
    return VdPciHdrPacket(
        rq=0 if not rq or rq in [False, RqBit.RESPONSE, 0] else 1, vendor_id=vendor_id, vdm_cmd_code=vdm_cmd_code
    )


class AutobindVDMMsg:
    def __init__(self, vid: VdPCIVendorIds, vdm_cmd_code):
        self.vdm_cmd_code = vdm_cmd_code
        self.vid = vid

    def __call__(self, cls: type[Packet]):
        vid = self.vid
        cmd_code = self.vdm_cmd_code
        bind_layers(
            VdPciHdrPacket,
            cls,
            vid=vid.value if type(vid) == VdPCIVendorIds else vid,
            vdm_cmd_code=cmd_code.value if hasattr(cmd_code, "value") else cmd_code,
        )
        if not hasattr(cls, "name") or cls.name is None:
            cls.name = cls.__name__
        if not hasattr(cls, "vid") or cls.vid is None:
            cls.vid = self.vid
        if not hasattr(cls, "vdm_cmd_code") or cls.vdm_cmd_code is None:
            cls.vdm_cmd_code = self.vdm_cmd_code
        return cls



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from typing import Any

from scapy.fields import BitEnumField, BitField, FieldLenField, FieldListField, PacketListField, XByteField
from scapy.packet import Packet

from ...helpers import AllowRawSummary
from .. import EndpointContext
from ..types import AnyPacketType, EntryType, RoutingTableEntry
from .control import (
    AutobindControlMsg,
    ControlHdr,
    ControlHdrPacket,
    RqBit,
    set_control_fields,
)
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


class RoutingTableEntryPacket(AllowRawSummary, Packet):
    name = "RoutingTableEntry"

    fields_desc = [
        XByteField("eid_range", 0),
        XByteField("starting_eid", 0),
        BitEnumField("entry_type", 0, 2, EntryType),
        BitField("static_eid", 0, 1),
        BitField("port_number", 0, 5),
        XByteField("phys_transport_binding_id", 0),
        XByteField("phy_media_type_id", 0),
        FieldLenField("phys_address_size", None, fmt="B", count_of="phy_address"),
        FieldListField("phy_address", [], XByteField("", 0), count_from=lambda pkt: pkt.phys_address_size),
    ]

    def extract_padding(self, p):
        """Required to ensure remaining bytes are properly transferred into next entry"""
        return b"", p

    def to_dict(self) -> dict[str, Any]:
        data = {}
        for f in self.fields_desc:
            value = getattr(self, f.name)
            if value is type(None):
                value = None
            data[f.name] = value
        return data


def NewRoutingTableEntry(entry: RoutingTableEntry) -> RoutingTableEntryPacket:
    return RoutingTableEntryPacket(**entry.to_dict())


@AutobindControlMsg(ContrlCmdCodes.GetRoutingTableEntries)
class GetRoutingTableEntriesPacket(AllowRawSummary, Packet):
    name = "GetRoutingTableEntries"

    fields_desc = set_control_fields(
        # TODO: define response fields
        rq_fields=[
            XByteField("entry_handle", 0),
        ],
        rsp_fields=[
            XByteField("next_entry_handle", 0),
            FieldLenField("entry_count", None, fmt="B", count_of="entries"),
            # TODO: implement routing table entry field
            PacketListField("entries", [], RoutingTableEntryPacket, count_from=lambda pkt: pkt.entry_count),
        ],
    )

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        # if we made it hear then the msg_type is unsupported
        # 0x80: message type number not supported
        if not ctx.routing_table_ready:
            return CompletionCodes.ERROR_NOT_READY, None

        rt_entries = [NewRoutingTableEntry(e) for e in ctx.routing_table]
        return CompletionCodes.SUCCESS, GetRoutingTableEntriesResponse(next_entry_handle=0xFF, entries=rt_entries)

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"{self.name} ("
        if self.underlayer.getfieldval("rq") == RqBit.REQUEST.value:
            summary += f"hdl=0x{self.entry_handle:02X})"
        else:
            summary += f"next_hdl=0x{self.next_entry_handle:02X}, cnt={self.entry_count}) "
            for entry in self.entries:
                summary += f" [0x{entry.starting_eid:02X}:{entry.eid_range}]"
        return summary, [ControlHdrPacket]


def GetRoutingTableEntries(_pkt: bytes | bytearray = b"", /, *, entry_handle: int = 0) -> GetRoutingTableEntriesPacket:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.GetRoutingTableEntries)
    if _pkt:
        return GetRoutingTableEntriesPacket(_pkt, _underlayer=hdr)
    return GetRoutingTableEntriesPacket(
        entry_handle=entry_handle,
        _underlayer=hdr,
    )


def GetRoutingTableEntriesResponse(
    _pkt: bytes | bytearray = b"", /, *, next_entry_handle: int, entries: list[int] | None = None
) -> GetRoutingTableEntriesPacket:
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.GetRoutingTableEntries)
    if _pkt:
        return GetRoutingTableEntriesPacket(_pkt, _underlayer=hdr)
    return GetRoutingTableEntriesPacket(
        next_entry_handle=next_entry_handle,
        entry_count=len(entries),
        entries=entries,
        # add a default underlayer to set the required "rq" field
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum

from scapy.fields import FieldListField, XByteEnumField, XByteField, XLEIntField
from scapy.packet import Packet

from ..transport import TransportHdrPacket
from ..types import AnyPacketType
from .pldm import AutobindPLDMMsg, PldmHdrPacket, set_pldm_fields
from .types import (
    PldmControlCmdCodes,
    PldmTypeCodes,
)


@AutobindPLDMMsg(PldmTypeCodes.CONTROL, PldmControlCmdCodes.SetTID)
class SetTIDPacket(Packet):
    fields_desc = set_pldm_fields(
        rq_fields=[
            XByteField("tid", 0),
        ],
        rsp_fields=[],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"SETTID (tid: {self.tid})"
        return summary, [PldmHdrPacket, TransportHdrPacket]


@AutobindPLDMMsg(PldmTypeCodes.CONTROL, PldmControlCmdCodes.GetTID)
class GetTIDPacket(Packet):
    fields_desc = set_pldm_fields(
        rq_fields=[],
        rsp_fields=[
            XByteField("tid", 0),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"GETTID (tid: {self.tid})"
        return summary, [PldmHdrPacket, TransportHdrPacket]


class GetPLDMVersionOperation(IntEnum):
    GET_NEXT_PART = 0
    GET_FIRST_PART = 1


class GetPLDMVersionTransferFlag(IntEnum):
    START = 1
    MIDDLE = 2
    END = 4
    START_AND_END = 5


@AutobindPLDMMsg(PldmTypeCodes.CONTROL, PldmControlCmdCodes.GetPLDMVersion)
class GetPLDMVersionPacket(Packet):
    fields_desc = set_pldm_fields(
        rq_fields=[
            XLEIntField("DataTransferHandle", 0),
            XByteEnumField("TransferOperationFlag", 0, GetPLDMVersionOperation),
            XByteField("PLDMType", 0),
        ],
        rsp_fields=[
            XLEIntField("NextDataTransferHandle", 0),
            XByteEnumField("TransferFlag", 0, GetPLDMVersionTransferFlag),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = "GetPLDMVER ("
        if self.underlayer.getfieldval("rq") == 1:
            summary += f"hdl={self.DataTransferHandle}, op={self.TransferOperationFlag}, type={self.PLDMType})"
        else:
            summary += f"nextHdl={self.NextDataTransferHandle}, transferFlag={self.TransferFlag})"
        return summary, [PldmHdrPacket, TransportHdrPacket]


class PLDMTypesByte1(IntEnum):
    CONTROL = 1 << 0x00
    SMBIOS = 1 << 0x01
    PLATFORM_MONITORING = 1 << 0x02
    BIOS = 1 << 0x03
    FRU = 1 << 0x04
    FIRMWARE_UPDATES = 1 << 0x05
    RDE = 1 << 0x06


@AutobindPLDMMsg(PldmTypeCodes.CONTROL, PldmControlCmdCodes.GetPLDMTypes)
class GetPLDMTypesPacket(Packet):
    fields_desc = set_pldm_fields(
        rq_fields=[],
        rsp_fields=[
            XByteEnumField("PLDMTypes1", 0, PLDMTypesByte1),
            XByteField("PLDMTypes2", 0),
            XByteField("PLDMTypes3", 0),
            XByteField("PLDMTypes4", 0),
            XByteField("PLDMTypes5", 0),
            XByteField("PLDMTypes6", 0),
            XByteField("PLDMTypes7", 0),
            XByteField("PLDMTypes8", 0),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = "GetPLDMTypes ("
        if self.underlayer.getfieldval("rq") == 1:
            summary += ")"
            return summary, [PldmHdrPacket, TransportHdrPacket]

        if self.PLDMTypes1 != 0:
            if self.PLDMTypes1 & PLDMTypesByte1.CONTROL != 0:
                summary += "CTRL "
            if self.PLDMTypes1 & PLDMTypesByte1.SMBIOS != 0:
                summary += "SMBIOS "
            if self.PLDMTypes1 & PLDMTypesByte1.PLATFORM_MONITORING != 0:
                summary += "M&C "
            if self.PLDMTypes1 & PLDMTypesByte1.BIOS != 0:
                summary += "BIOS "
            if self.PLDMTypes1 & PLDMTypesByte1.FRU != 0:
                summary += "FRU "
            if self.PLDMTypes1 & PLDMTypesByte1.FIRMWARE_UPDATES != 0:
                summary += "FUP "
            if self.PLDMTypes1 & PLDMTypesByte1.RDE != 0:
                summary += "RDE "
            summary = summary.rstrip()
        if self.PLDMTypes2 != 0:
            summary += f"PLDMTypes[2]=0x{self.PLDMTypes2:02X}"
        if self.PLDMTypes3 != 0:
            summary += f"PLDMTypes[3]=0x{self.PLDMTypes3:02X}"
        if self.PLDMTypes4 != 0:
            summary += f"PLDMTypes[4]=0x{self.PLDMTypes4:02X}"
        if self.PLDMTypes5 != 0:
            summary += f"PLDMTypes[5]=0x{self.PLDMTypes5:02X}"
        if self.PLDMTypes6 != 0:
            summary += f"PLDMTypes[6]=0x{self.PLDMTypes6:02X}"
        if self.PLDMTypes7 != 0:
            summary += f"PLDMTypes[7]=0x{self.PLDMTypes7:02X}"
        if self.PLDMTypes8 != 0:
            summary += f"PLDMTypes[8]=0x{self.PLDMTypes8:02X}"

        summary += ")"
        return summary, [PldmHdrPacket, TransportHdrPacket]


@AutobindPLDMMsg(PldmTypeCodes.CONTROL, PldmControlCmdCodes.GetPLDMCommands)
class GetPLDMCommandsPacket(Packet):
    fields_desc = set_pldm_fields(
        rq_fields=[
            XByteField("PLDMType", 0),
            XLEIntField("Version", 0),
        ],
        rsp_fields=[
            FieldListField("cmds", [], XByteField("", 0), length_from=lambda pkt: 32),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = "GetPLDMTypes ("
        if self.underlayer.getfieldval("rq") == 0:
            summary += f"{self.cmds}"
        summary += ")"
        return summary, [PldmHdrPacket, TransportHdrPacket]



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from scapy.fields import FieldLenField, FieldListField, XByteField, XLEIntField
from scapy.packet import Packet

from .. import EndpointContext, TransportHdrPacket
from ..types import AnyPacketType
from .control import (
    AutobindControlMsg,
    ControlHdr,
    ControlHdrPacket,
    set_control_fields,
)
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes

MCTP_BASE_SPEC_VERSION_1_3_1 = 0xF1F3F100
MCTP_CONTROL_VERSION_1_3_1 = 0xF1F3F100


@AutobindControlMsg(ContrlCmdCodes.GetMCTPVersionSupport)
class GetMctpVersionSupportPacket(Packet):
    name = "GetMctpVersionSupport"

    fields_desc = set_control_fields(
        rq_fields=[
            XByteField("msg_type_number", 0),
        ],
        rsp_fields=[
            FieldLenField("version_number_entry_count", None, fmt="!B", count_of="version_number_list"),
            FieldListField(
                "version_number_list", [], XLEIntField("", 0), count_from=lambda pkt: pkt.version_number_entry_count
            ),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"{self.name} ("
        if self.underlayer.getfieldval("rq") == 0:
            if self.version_number_entry_count != 0:
                summary += ", ".join(f"0x{x:08X}" for x in self.version_number_list)
        else:
            summary += f"msg_type_number: {self.msg_type_number}"
        summary += ")"
        return summary, [ControlHdrPacket, TransportHdrPacket]

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        msg_type_number = self.msg_type_number

        version_numbers = []
        if msg_type_number == 0:
            version_numbers += [MCTP_CONTROL_VERSION_1_3_1]
        elif msg_type_number == 0xFF:
            version_numbers += [MCTP_BASE_SPEC_VERSION_1_3_1]

        if version_numbers:
            return CompletionCodes.SUCCESS, GetMctpVersionSupportPacket(
                version_number_list=version_numbers,
            )

        # if we made it hear then the msg_type is unsupported
        # 0x80: message type number not supported
        return 0x80, None


def GetMctpVersionSupport(_pkt: bytes | bytearray = b"", /, *, msg_type_number: int = 0) -> GetMctpVersionSupportPacket:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.GetMCTPVersionSupport)
    if _pkt:
        return GetMctpVersionSupportPacket(_pkt, _underlayer=hdr)
    return GetMctpVersionSupportPacket(
        msg_type_number=msg_type_number,
        _underlayer=hdr,
    )


def GetMctpVersionSupportResponse(
    _pkt: bytes | bytearray = b"", /, *, version_number_list: list[int] | None = None
) -> GetMctpVersionSupportPacket:
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.GetMCTPVersionSupport)
    if _pkt:
        return GetMctpVersionSupportPacket(_pkt, _underlayer=hdr)
    return GetMctpVersionSupportPacket(
        msg_type_cnt=len(version_number_list),
        version_number_list=version_number_list,
        # add a default underlayer to set the required "rq" field
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum

from scapy.fields import BitEnumField, BitField, XByteField
from scapy.packet import Packet

from .. import EndpointContext, TransportHdrPacket
from ..types import AnyPacketType
from .control import (
    AutobindControlMsg,
    ControlHdr,
    ControlHdrPacket,
    set_control_fields,
)
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


class EndpointType(IntEnum):
    SIMPLE = 0
    """Simple Endpoint"""

    BUS_OWNER = 1
    """Bus Owner and/or Bridge"""


class EndpointIDType(IntEnum):
    DYNAMIC = 0
    """The endpoint uses a dynamic EID only"""

    STATIC_EID_SUPPORTED = 1
    """The EID returned by this command reflects the present setting and may or
    may not match the static EID value."""

    STATIC_EID_MATCH = 2
    """The endpoint has been configured with a static EID. The present value is
    the same as the static value."""

    STATIC_EID_MISMATCH = 3
    """ Endpoint has been configured with a static EID. The present value is
    different than the static value"""


@AutobindControlMsg(ContrlCmdCodes.GetEndpointID)
class GetEndpointIDPacket(Packet):
    name = "GetEndpointID"

    fields_desc = set_control_fields(
        rsp_fields=[
            XByteField("eid", 0),
            BitField("unused", 0, 2),
            BitEnumField("endpoint_type", 0, 2, EndpointType),
            BitField("unused2", 0, 2),
            BitEnumField("endpoint_id_type", 0, 2, EndpointIDType),
            XByteField("medium_specific", 0),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"{self.name}"
        if self.underlayer.getfieldval("rq") == 0:
            summary += f" (eid: {self.eid:02X}, type: "
            summary += "simple, " if self.endpoint_type == 0 else "busowner, "
            if self.endpoint_id_type == EndpointIDType.DYNAMIC.value:
                summary += "eid_type: dynamic"
            elif self.endpoint_id_type == EndpointIDType.STATIC_EID_SUPPORTED.value:
                summary += "eid_type: static_eid_supported"
            elif self.endpoint_id_type == EndpointIDType.STATIC_EID_MATCH.value:
                summary += "eid_type: static_eid_match"
            elif self.endpoint_id_type == EndpointIDType.STATIC_EID_MISMATCH.value:
                summary += "eid_type: static_eid_mismatch"
            summary += ")"
        return summary, [ControlHdrPacket, TransportHdrPacket]

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        cmplt_code = CompletionCodes.SUCCESS
        endp_type = EndpointIDType.DYNAMIC
        if ctx.static_eid:
            endp_type = EndpointIDType.STATIC_EID_SUPPORTED
            if ctx.assigned_eid:
                if ctx.assigned_eid == ctx.static_eid:
                    endp_type = EndpointIDType.STATIC_EID_MATCH
                else:
                    endp_type = EndpointIDType.STATIC_EID_MISMATCH
        return cmplt_code, GetEndpointIDResponse(
            eid=ctx.eid,
            endpoint_type=EndpointType.BUS_OWNER if ctx.is_bus_owner else EndpointType.SIMPLE,
            endpoint_id_type=endp_type,
        )


def GetEndpointID(*args) -> GetEndpointIDPacket:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.GetEndpointID)
    if len(args):
        return GetEndpointIDPacket(*args, _underlayer=hdr)
    return GetEndpointIDPacket(_underlayer=hdr)


def GetEndpointIDResponse(
    *args,
    eid: int = 0,
    endpoint_type: EndpointType = EndpointType.SIMPLE,
    endpoint_id_type: EndpointIDType = EndpointIDType.DYNAMIC,
    medium_specific: int = 0,
) -> GetEndpointIDPacket:
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.GetEndpointID)
    if len(args):
        return GetEndpointIDPacket(*args, _underlayer=hdr)
    return GetEndpointIDPacket(
        eid=eid,
        endpoint_type=endpoint_type,
        endpoint_id_type=endpoint_id_type,
        medium_specific=medium_specific,
        # add a default underlayer to set the required "rq" field
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum
from typing import Protocol, TypeVar, runtime_checkable

from ..types import AnyPacketType, EndpointContext


class ContrlCmdCodes(IntEnum):
    Reserved = 0x00
    SetEndpointID = 0x01
    GetEndpointID = 0x02
    GetEndpointUUID = 0x03
    GetMCTPVersionSupport = 0x04
    GetMessageTypeSupport = 0x05
    GetVendorDefinedMessageSupport = 0x06
    ResolveEndpointID = 0x07
    AllocateEndpointIDs = 0x08
    RoutingInformationUpdate = 0x09
    GetRoutingTableEntries = 0x0A
    PrepareForEndpointDiscovery = 0x0B
    EndpointDiscovery = 0x0C
    DiscoveryNotify = 0x0D
    GetNetworkID = 0x0E
    QueryHop = 0x0F
    ResolveUUID = 0x10
    QueryRateLimit = 0x11
    RequestTXRateLimit = 0x12
    UpdateRateLimit = 0x13
    QuerySupportedInterfaces = 0x14
    Unknown = 0xFF


class CompletionCodes(IntEnum):
    SUCCESS = 0
    """The Request was accepted and completed normally."""

    ERROR = 1
    """This is a generic failure message. (It should not be used
    when a more specific result code applies.)"""

    ERROR_INVALID_DATA = 2
    """The packet payload contained invalid data or an illegal
    parameter value."""

    ERROR_INVALID_LENGTH = 3
    """The message length was invalid. (The Message body
    was larger or smaller than expected for the particular
    request.)"""

    ERROR_NOT_READY = 4
    """The Receiver is in a transient state where it is not ready
    to receive the corresponding message."""

    ERROR_UNSUPPORTED_CMD = 5
    """The command field in the control type of the received
    message is unspecified or not supported on this
    endpoint. This completion code shall be returned for any
    unsupported command values received in MCTP control
    Request messages."""


CompletionCode = TypeVar("CompletionCode", CompletionCodes, int)


@runtime_checkable
class IControlMsgCanReply(Protocol):
    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        pass


@runtime_checkable
class IControlMsgPacket(Protocol):
    cmd_code: ContrlCmdCodes



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import time
from collections.abc import Callable
from enum import IntEnum

from scapy.config import conf
from scapy.fields import AnyField, BitEnumField, BitField, ConditionalField, XBitField, XByteField
from scapy.packet import Packet, Raw, bind_layers

from ...interfaces import ICanSetMySummaryClasses
from .. import EndpointContext
from ..transport import AutobindMessageType, MsgTypes, SmbusTransportPacket, TransportHdrPacket
from ..types import AnyPacketType
from .types import CompletionCodes, PldmTypeCodes


class RqBit(IntEnum):
    RESPONSE = 0
    REQUEST = 1


DEFAULT_HDR_VERSION = 0


@AutobindMessageType(MsgTypes.PLDM)
class PldmHdrPacket(Packet):
    name = "PLDM"
    fields_desc = [
        BitEnumField("rq", 0, 1, RqBit),
        BitField("d", 0, 1),
        BitField("unused", 0, 1),
        XBitField("instance_id", 0, 5),
        BitField("hdr_ver", DEFAULT_HDR_VERSION, 2),
        BitEnumField("pldm_type", 0, 6, PldmTypeCodes),
        XByteField("cmd_code", 0),
        ConditionalField(XByteField("completion_code", 0), lambda pkt: pkt.rq == RqBit.RESPONSE.value),
    ]

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        rqType = "REQ" if self.rq == RqBit.REQUEST.value else "RSP"
        summary = (
            f"PLDM {rqType} (inst_id=0x{self.instance_id:02X}, type=0x{self.pldm_type:02X}, cmd=0x{self.cmd_code:02X}"
        )
        if self.rq == RqBit.RESPONSE.value:
            summary += f", cc={self.completion_code})"
        else:
            summary += ")"
        return summary, [TransportHdrPacket, SmbusTransportPacket]

    def do_dissect_payload(self, s: bytes) -> None:
        if not s:
            return
        cls = self.guess_payload_class(s)
        try:
            p = cls(s, _internal=1, _underlayer=self)
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector and cls is not None:
                raise
            p = conf.raw_layer(s, _internal=1, _underlayer=self)
        self.add_payload(p)
        if isinstance(p, ICanSetMySummaryClasses):
            p.set_mysummary_classes([PldmHdrPacket, TransportHdrPacket])

    def answers(self, other: Packet) -> int:
        if not other.haslayer(PldmHdrPacket):
            return 0
        if self.rq != 0 or other.rq != 1:
            return 0
        if self.pldm_type != other.pldm_type:
            return 0
        if self.cmd_code != other.cmd_code:
            return 0
        if self.hdr_ver != DEFAULT_HDR_VERSION:
            print(f"WARN: Mismatched header versions: {self.hdr_ver} != {DEFAULT_HDR_VERSION}")
        return self.payload.answers(other.payload)

    def is_request(self, check_payload: bool = True) -> bool:
        return self.rq == 1

    def make_reply(self, ctx: EndpointContext) -> AnyPacketType:
        if not self.is_request():
            return None
        if self.pldm_type not in list(PldmTypeCodes):
            return None
        # only make a reply if we are a supported msg type (allows context to control which packets generate responses)
        if MsgTypes.PLDM not in ctx.supported_msg_types:
            return None

        completion_code = CompletionCodes.ERROR_UNSUPPORTED_CMD
        payload_resp = None

        # Check if we have a premade response
        if ctx.mctp_responses:
            pldm_hdr: PldmHdrPacket = self.getlayer(PldmHdrPacket)
            data = bytes(pldm_hdr)
            rqDInstanceID, req_data = data[0], data[1:]
            resp_info = ctx.get_response(MsgTypes.PLDM, req_data)
            if resp_info:
                print(f"***> PLDM Request  ({len(resp_info.data)} bytes): {resp_info.description}")
                resp_data = Raw(resp_info.data)

                delay = resp_info.processing_delay
                if delay:
                    time.sleep(delay / 1000.0)
                return Raw(bytes([rqDInstanceID & 0x7F])) / resp_data
        # TODO: add support to auto-respond to pldm msgs

        rsp = PldmHdr(
            rq=False,
            instance_id=self.instance_id,
            pldm_type=self.pldm_type,
            hdr_ver=self.hdr_ver,
            cmd_code=self.cmd_code,
            completion_code=completion_code,
        )
        return (rsp / payload_resp) if payload_resp else rsp


def PldmHdr(
    *args,
    rq: bool | RqBit = RqBit.RESPONSE,
    d: bool = False,
    instance_id: int = 0,
    hdr_ver: int = DEFAULT_HDR_VERSION,
    pldm_type: PldmTypeCodes = PldmTypeCodes.CONTROL,
    cmd_code: int = 0,
    completion_code: int | None = 0,
) -> PldmHdrPacket:
    if len(args):
        return PldmHdrPacket(*args)
    return PldmHdrPacket(
        rq=0 if not rq or rq in [False, RqBit.RESPONSE, 0] else 1,
        d=1 if d else 0,
        instance_id=instance_id,
        hdr_ver=hdr_ver,
        pldm_type=pldm_type,
        cmd_code=cmd_code,
        completion_code=completion_code,
    )


class AutobindPLDMMsg:
    def __init__(self, pldm_type: PldmTypeCodes, cmd_code: int):
        self.pldm_type = pldm_type
        self.cmd_code = cmd_code

    def __call__(self, cls: type[Packet]):
        pldm_type = self.pldm_type
        cmd_code = self.cmd_code
        # print(f"Binding cls {cls} to cmd_code {cmd_code}:{self.is_request}")
        bind_layers(
            PldmHdrPacket,
            cls,
            pldm_type=pldm_type.value if type(pldm_type) == PldmTypeCodes else pldm_type,
            cmd_code=cmd_code.value if hasattr(cmd_code, "value") else cmd_code,
        )
        if not hasattr(cls, "name") or cls.name is None:
            cls.name = cls.__name__
        if not hasattr(cls, "pldm_type") or cls.pldm_type is None:
            cls.pldm_type = self.pldm_type
        if not hasattr(cls, "cmd_code") or cls.cmd_code is None:
            cls.cmd_code = self.cmd_code
        return cls


def set_pldm_fields(
    rq_fields: list[AnyField] | None = None, rsp_fields: list[AnyField] | None = None
) -> list[AnyField]:
    rq_fields = rq_fields or []
    rsp_fields = rsp_fields or []

    def gen_conditional_field(fld: AnyField, cond: Callable[[Packet], bool]):
        def default_cond(pkt):
            return True

        if isinstance(fld, ConditionalField):
            # unwrap the conditional field
            default_cond = fld.cond
            fld = fld.fld
        return ConditionalField(fld=fld, cond=lambda pkt: all([cond(pkt), default_cond(pkt)]))

    def is_request(pkt: Packet) -> bool:
        return pkt.underlayer.getfieldval("rq") == RqBit.REQUEST.value

    def is_response(pkt: Packet) -> bool:
        return pkt.underlayer.getfieldval("rq") == RqBit.RESPONSE.value

    fields = [gen_conditional_field(fld, is_request) for fld in rq_fields]
    fields += [gen_conditional_field(fld, is_response) for fld in rsp_fields]

    return fields



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from scapy.packet import Packet

from .. import EndpointContext
from ..types import AnyPacketType
from .control import AutobindControlMsg, ControlHdr
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


@AutobindControlMsg(ContrlCmdCodes.DiscoveryNotify)
class DiscoveryNotifyPacket(Packet):
    fields_desc = []

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        return CompletionCodes.SUCCESS, DiscoveryNotifyResponse()


def DiscoveryNotify(*args, **kwargs):
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.DiscoveryNotify)
    if len(args):
        return DiscoveryNotifyPacket(*args, _underlayer=hdr)
    return DiscoveryNotifyPacket(
        _underlayer=hdr,
    )


def DiscoveryNotifyResponse(*args, **kwargs):
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.DiscoveryNotify)
    if len(args) or len(kwargs):
        return DiscoveryNotifyPacket(*args, _underlayer=hdr, **kwargs)
    return DiscoveryNotifyPacket(
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

from enum import IntEnum

from scapy.fields import BitEnumField, BitField, XByteField
from scapy.packet import Packet

from .. import EndpointContext, TransportHdrPacket
from ..types import AnyPacketType
from .control import AutobindControlMsg, ControlHdr, ControlHdrPacket, set_control_fields
from .types import CompletionCode, CompletionCodes, ContrlCmdCodes


class SetEndpointIDOperation(IntEnum):
    SetEID = 0
    """Submit an EID for assignment. The given EID will be accepted conditional upon which bus the device received
    the EID from (see preceding text). A device where the endpoint is only reached through one bus shall always
    accept this operation (provided the EID value is legal)."""

    ForceEID = 1
    """Force EID assignment. The given EID will be accepted regardless of whether the EID was already assigned
    through another bus. Note that if the endpoint is forcing, the EID assignment changes which bus is being tracked
    as the originator of the Set Endpoint ID command. A device where the endpoint is only reached through one bus
    shall always accept this operation (provided the EID value is legal), in which case the Set EID and Force EID
    operations are equivalent."""

    ResetEID = 2
    """This option only applies to endpoints that support static EIDs. If static EIDs are supported, the endpoint
    shall restore the EID to the statically configured EID value. The EID value in byte 2 shall be ignored. An
    ERROR_INVALID_DATA completion code shall be returned if this operation is not supported."""

    SetDiscoveredFlag = 3
    """Set Discovered flag to the “discovered” state only. Do not change present EID setting. The EID value in byte 2
    shall be ignored.

    Note that Discovered flag is only used for some physical transport bindings. An
    ERROR_INVALID_DATA completion code shall be returned if this operation is selected and the particular transport
    binding does not support a Discovered flag."""


class SetEndpointIDAssignmentStatus(IntEnum):
    ACCEPTED = 0
    REJECTED = 1


class SetEndpointIDAllocationStatus(IntEnum):
    NO_EID_POOL_REQUIRED = 0
    EID_POOL_REQUIRED = 1
    EID_POOL_ALREADY_ASSIGNED = 2


@AutobindControlMsg(ContrlCmdCodes.SetEndpointID)
class SetEndpointIDPacket(Packet):
    fields_desc = set_control_fields(
        rq_fields=[
            BitField("reserved1", 0, 6),
            BitEnumField("op", 0, 2, SetEndpointIDOperation),
            XByteField("eid", 0),
        ],
        rsp_fields=[
            BitField("reserved2", 0, 2),
            BitEnumField("eid_assignment_status", 0, 2, SetEndpointIDAssignmentStatus),
            BitField("reserved3", 0, 2),
            BitEnumField("eid_allocation_status", 0, 2, SetEndpointIDAllocationStatus),
            XByteField("eid_setting", 0),
            XByteField("eid_pool_size", 0),
        ],
    )

    def mysummary(self) -> str | tuple[str, list[AnyPacketType]]:
        summary = f"{self.name} ("
        if self.underlayer.getfieldval("rq") == 0:
            summary += "assign_status: "
            summary += "accepted, " if self.eid_assignment_status == 0 else "rejected, "

            summary += "eid_alloc_status: "
            if self.eid_allocation_status == SetEndpointIDAllocationStatus.NO_EID_POOL_REQUIRED.value:
                summary += "no_pool, "
            elif self.eid_allocation_status == SetEndpointIDAllocationStatus.EID_POOL_REQUIRED.value:
                summary += "pool_required, "
            elif self.eid_allocation_status == SetEndpointIDAllocationStatus.EID_POOL_ALREADY_ASSIGNED.value:
                summary += "pool_assigned, "

            summary += f"eid_setting: 0x{self.eid_setting:02X}, eid_pool_size: {self.eid_pool_size}"
        else:
            summary += f"eid: 0x{self.eid:02X}, op: "
            if self.op == SetEndpointIDOperation.SetEID.value:
                summary += "set"
            elif self.op == SetEndpointIDOperation.ForceEID.value:
                summary += "force"
            elif self.op == SetEndpointIDOperation.ResetEID.value:
                summary += "reset"
            elif self.op == SetEndpointIDOperation.SetDiscoveredFlag.value:
                summary += "set_disc"
        summary += ")"
        return summary, [ControlHdrPacket, TransportHdrPacket]

    def make_ctrl_reply(self, ctx: EndpointContext) -> tuple[CompletionCode, AnyPacketType]:
        cmplt_code = CompletionCodes.SUCCESS
        op: SetEndpointIDOperation = self.op
        eid: int = self.eid

        assignment_status = SetEndpointIDAssignmentStatus.REJECTED
        if op in (SetEndpointIDOperation.SetEID, SetEndpointIDOperation.ForceEID):
            ctx.assigned_eid = eid
            ctx.discovered = True
            assignment_status = SetEndpointIDAssignmentStatus.ACCEPTED
        elif op == SetEndpointIDOperation.SetDiscoveredFlag:
            ctx.discovered = True
        elif op == SetEndpointIDOperation.ResetEID:
            ctx.assigned_eid = ctx.static_eid
        else:
            return CompletionCodes.ERROR_INVALID_DATA, None

        alloc_status = SetEndpointIDAllocationStatus.NO_EID_POOL_REQUIRED
        if ctx.pool_size:
            if ctx.allocated_pool:
                alloc_status = SetEndpointIDAllocationStatus.EID_POOL_ALREADY_ASSIGNED
            else:
                alloc_status = SetEndpointIDAllocationStatus.EID_POOL_REQUIRED

        return cmplt_code, SetEndpointIDResponse(
            eid_allocation_status=alloc_status,
            eid_pool_size=ctx.pool_size,
            eid_assignment_status=assignment_status,
            eid_setting=ctx.assigned_eid,
        )


def SetEndpointID(*args, op: SetEndpointIDOperation, eid: int) -> Packet:
    hdr = ControlHdr(rq=True, cmd_code=ContrlCmdCodes.SetEndpointID)
    if len(args):
        return SetEndpointIDPacket(*args, _underlayer=hdr)
    return SetEndpointIDPacket(
        op=op,
        eid=eid,
        _underlayer=hdr,
    )


def SetEndpointIDResponse(
    *args,
    eid_assignment_status: SetEndpointIDAssignmentStatus,
    eid_allocation_status: SetEndpointIDAllocationStatus,
    eid_setting: int,
    eid_pool_size: int,
) -> Packet:
    hdr = ControlHdr(rq=False, cmd_code=ContrlCmdCodes.SetEndpointID)
    if len(args):
        return SetEndpointIDPacket(*args, _underlayer=hdr)
    return SetEndpointIDPacket(
        eid_assignment_status=eid_assignment_status,
        eid_allocation_status=eid_allocation_status,
        eid_setting=eid_setting,
        eid_pool_size=eid_pool_size,
        # add a default underlayer to set the required "rq" field
        _underlayer=hdr,
    )



# SPDX-FileCopyrightText: 2024 Justin Simon <justin@simonctl.com>
#
# SPDX-License-Identifier: MIT

import contextlib
import socket
import time

from scapy.compat import raw
from scapy.config import conf
from scapy.data import MTU
from scapy.interfaces import _GlobInterfaceType, network_name
from scapy.packet import Packet
from scapy.plist import PacketList, SndRcvList
from scapy.sendrecv import sndrcv
from scapy.supersocket import SuperSocket
from scapy.utils import linehexdump

from pymctp.layers.mctp import SmbusTransport


class QemuI2CNetDevSocket(SuperSocket):
    desc = "read/write to a Qemu NetDev Socket"

    def __init__(
        self,
        family: int = socket.AF_INET,
        type: int = socket.SOCK_DGRAM,  # noqa: A002
        proto: int = 0,
        iface: _GlobInterfaceType | None = None,
        in_port=0,
        out_port=None,
        id_str="",
        dump_hex=True,
        dump_packet=False,
        **kwargs,
    ):
        self.id_str = id_str
        self.dump_hex = dump_hex
        self.dump_packet = dump_packet
        fd = socket.socket(family, type, proto)
        assert fd != -1
        self.ins = self.outs = fd

        self.ins.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if iface is not None:
            iface2 = network_name(iface)
            self.iface = iface2
        else:
            self.iface = "any"

        self.ins.bind((iface, in_port))
        if out_port:
            # self.outs.connect((iface, out_port))
            self.out_port = out_port

    def send(self, x: Packet) -> int:
        """Overloaded Packet.send() method to use 'socket.sendto' to connect to the address
        before sending the packet.
        """
        sx = raw(x)
        with contextlib.suppress(AttributeError):
            x.sent_time = time.time()

        if not self.outs:
            return 0

        if self.out_port:
            try:
                result = self.outs.sendto(sx, (self.iface, self.out_port))
            except Exception as e:
                print(f"Failed sending data: {e}")
                raise
            else:
                if self.dump_hex:
                    print(f"{self.id_str}>TX> {linehexdump(x, onlyhex=1, dump=True)}")
                if self.dump_packet:
                    print(f"{self.id_str}>TX> {x.summary()}")
                return result
        else:
            return self.outs.send(sx)

    def recv(self, x: int = MTU) -> Packet | None:
        raw_bytes = self.ins.recv(x)
        if self.dump_hex:
            print(f"{self.id_str}<RX< {linehexdump(raw_bytes, onlyhex=1, dump=True)}")
        if len(raw_bytes) < 7:
            return None
        # TODO: Move this to a config field to support multiple transports
        #       Not needed right now as Qemu only supports I2C/SMBUS payloads
        pkt = SmbusTransport(raw_bytes)
        pkt.time = time.time()
        if pkt and self.dump_packet:
            print(f"{self.id_str}<RX< {pkt.summary()}")
        return pkt


# @conf.commands.register
# def srqemu(address, pkts, inter=0.1, *args, in_port=0, out_port=None, **kwargs) -> tuple[SndRcvList, PacketList]:
#     """Send and receive using a QEMU I2C socket"""
#     s = QemuI2CNetDevSocket(iface=address, in_port=in_port, out_port=out_port)
#     a, b = sndrcv(s, pkts, inter=inter, *args, **kwargs)
#     s.close()
#     return a, b


# @conf.commands.register
# def srqemu1(address, pkts, inter=0.1, *args, in_port=0, out_port=None, **kwargs) -> Packet:
#     """Send and receive 1 packet using a QEMU I2C socket"""
#     a, b = srqemu(address, pkts, inter=inter, *args, in_port=in_port, out_port=out_port, **kwargs)
#     if len(a) > 0:
#         return a[0][1]
#     return None
